from typing import Any, Dict, List, Optional, Tuple

import cv2
import depthai as dai
import numpy as np

from depthai_nodes.message.creators import create_detection_message
from depthai_nodes.node.parsers.base_parser import BaseParser
from depthai_nodes.node.parsers.utils import (
    normalize_bboxes,
    xyxy_to_xywh,
)
from depthai_nodes.node.parsers.utils.masks_utils import (
    get_segmentation_outputs,
    process_single_mask,
)
from depthai_nodes.node.parsers.utils.yolo import (
    YOLOSubtype,
    decode_yolo_output,
    parse_kpts,
)


class YOLOExtendedParser(BaseParser):
    """Parser class for parsing the output of the YOLO Instance Segmentation and Pose
    Estimation models.

    Attributes
    ----------
    conf_threshold : float
        Confidence score threshold for detected faces.
    n_classes : int
        Number of classes in the model.
    label_names : Optional[List[str]]
        Names of the classes.
    iou_threshold : float
        Intersection over union threshold.
    mask_conf : float
        Mask confidence threshold.
    n_keypoints : int
        Number of keypoints in the model.
    anchors : Optional[List[List[List[float]]]]
        Anchors for the YOLO model (optional).
    keypoint_label_names : Optional[List[str]]
        Labels for the keypoints.
    keypoint_edges : Optional[List[Tuple[int, int]]]
        Keypoint connection pairs for visualizing the skeleton. Example: [(0,1), (1,2), (2,3), (3,0)] shows that keypoint 0 is connected to keypoint 1, keypoint 1 is connected to keypoint 2, etc.
    subtype : str
        Version of the YOLO model.


    Output Message/s
    ----------------
    **Type**: ImgDetectionsExtended

    **Description**: Message containing bounding boxes, labels, label names, confidence scores, and keypoints or masks and protos of the detected objects.
    """

    _DET_MODE = 0
    _KPTS_MODE = 1
    _SEG_MODE = 2

    def __init__(
        self,
        conf_threshold: float = 0.5,
        n_classes: int = 1,
        label_names: Optional[List[str]] = None,
        iou_threshold: float = 0.5,
        mask_conf: float = 0.5,
        n_keypoints: int = 17,
        anchors: Optional[List[List[List[float]]]] = None,
        subtype: str = "",
        keypoint_label_names: Optional[List[str]] = None,
        keypoint_edges: Optional[List[Tuple[int, int]]] = None,
    ):
        """Initialize the parser node.

        @param conf_threshold: The confidence threshold for the detections
        @type conf_threshold: float
        @param n_classes: The number of classes in the model
        @type n_classes: int
        @param label_names: The names of the classes
        @type label_names: Optional[List[str]]
        @param iou_threshold: The intersection over union threshold
        @type iou_threshold: float
        @param mask_conf: The mask confidence threshold
        @type mask_conf: float
        @param n_keypoints: The number of keypoints in the model
        @type n_keypoints: int
        @param anchors: The anchors for the YOLO model
        @type anchors: Optional[List[List[List[float]]]]
        @param subtype: The version of the YOLO model
        @type subtype: str
        @param keypoint_label_names: The labels for the keypoints
        @type keypoint_label_names: Optional[List[str]]
        @param keypoint_edges: Connection pairs of the keypoints. Example: [(0,1),
            (1,2), (2,3), (3,0)] shows that keypoint 0 is connected to keypoint 1,
            keypoint 1 is connected to keypoint 2, etc.
        @type keypoint_edges: Optional[List[Tuple[int, int]]]
        """
        super().__init__()

        self.output_layer_names = []
        self.conf_threshold = conf_threshold
        self.n_classes = n_classes
        self.label_names = label_names
        self.iou_threshold = iou_threshold
        self.mask_conf = mask_conf
        self.n_keypoints = n_keypoints
        self.anchors = anchors
        self.keypoint_label_names = keypoint_label_names
        self.keypoint_edges = keypoint_edges
        try:
            self.subtype = YOLOSubtype(subtype.lower())
        except ValueError as err:
            raise ValueError(
                f"Invalid YOLO version {subtype}. Supported YOLO versions are {[e.value for e in YOLOSubtype][:-1]}."
            ) from err

        self._logger.debug(
            f"YOLOExtendedParser initialized with conf_threshold={self.conf_threshold}, n_classes={self.n_classes}, label_names={self.label_names}, iou_threshold={self.iou_threshold}, mask_conf={self.mask_conf}, n_keypoints={self.n_keypoints}, anchors={self.anchors}, subtype='{self.subtype}', keypoint_label_names={self.keypoint_label_names}, keypoint_edges={self.keypoint_edges}"
        )

    def setOutputLayerNames(self, output_layer_names: List[str]) -> None:
        """Sets the output layer names for the parser.

        @param output_layer_names: The output layer names for the parser.
        @type output_layer_names: List[str]
        """
        if not isinstance(output_layer_names, list):
            raise ValueError("Output layer names must be a list.")
        if not all(isinstance(layer, str) for layer in output_layer_names):
            raise ValueError("Output layer names must be a list of strings.")

        self.output_layer_names = output_layer_names
        self._logger.debug(f"Output layer names set to {self.output_layer_names}")

    def setConfidenceThreshold(self, threshold: float) -> None:
        """Sets the confidence score threshold for detected faces.

        @param threshold: Confidence score threshold for detected faces.
        @type threshold: float
        """
        if not isinstance(threshold, float):
            raise ValueError("Confidence score threshold must be a float.")

        self.conf_threshold = threshold
        self._logger.debug(f"Confidence score threshold set to {self.conf_threshold}")

    def setNumClasses(self, n_classes: int) -> None:
        """Sets the number of classes in the model.

        @param numClasses: The number of classes in the model.
        @type numClasses: int
        """
        if not isinstance(n_classes, int):
            raise ValueError("Number of classes must be an integer.")
        if n_classes < 1:
            raise ValueError("Number of classes must be greater than 0.")

        self.n_classes = n_classes
        self._logger.debug(f"Number of classes set to {self.n_classes}")

    def setIouThreshold(self, iou_threshold: float) -> None:
        """Sets the intersection over union threshold.

        @param iou_threshold: The intersection over union threshold.
        @type iou_threshold: float
        """
        if not isinstance(iou_threshold, float):
            raise ValueError("Intersection over union threshold must be a float.")
        if iou_threshold < 0 or iou_threshold > 1:
            raise ValueError(
                "Intersection over union threshold must be between 0 and 1."
            )

        self.iou_threshold = iou_threshold
        self._logger.debug(
            f"Intersection over union threshold set to {self.iou_threshold}"
        )

    def setMaskConfidence(self, mask_conf: float) -> None:
        """Sets the mask confidence threshold.

        @param mask_conf: The mask confidence threshold.
        @type mask_conf: float
        """
        if not isinstance(mask_conf, float):
            raise ValueError("Mask confidence threshold must be a float.")
        if mask_conf < 0 or mask_conf > 1:
            raise ValueError("Mask confidence threshold must be between 0 and 1.")

        self.mask_conf = mask_conf
        self._logger.debug(f"Mask confidence threshold set to {self.mask_conf}")

    def setNumKeypoints(self, n_keypoints: int) -> None:
        """Sets the number of keypoints in the model.

        @param n_keypoints: The number of keypoints in the model.
        @type n_keypoints: int
        """
        if not isinstance(n_keypoints, int):
            raise ValueError("Number of keypoints must be an integer.")
        if n_keypoints < 1:
            raise ValueError("Number of keypoints must be greater than 0.")

        self.n_keypoints = n_keypoints
        self._logger.debug(f"Number of keypoints set to {self.n_keypoints}")

    def setAnchors(self, anchors: List[List[List[float]]]) -> None:
        """Sets the anchors for the YOLO model.

        @param anchors: The anchors for the YOLO model.
        @type anchors: List[List[List[float]]]
        """
        for anchor in anchors:
            if not isinstance(anchor, list):
                raise ValueError("Anchors must be a list of lists.")
            if not all(isinstance(val, list) for val in anchor):
                raise ValueError("Anchors must be a list of lists of lists.")
            if not all(isinstance(val, float) for sublist in anchor for val in sublist):
                raise ValueError("Anchors must be a list of lists of floats.")

        self.anchors = anchors
        self._logger.debug(f"Anchors set to {self.anchors}")

    def setSubtype(self, subtype: str) -> None:
        """Sets the subtype of the YOLO model.

        @param subtype: The subtype of the YOLO model.
        @type subtype: YOLOSubtype
        """
        if not isinstance(subtype, str):
            raise ValueError("Subtype must be a string.")

        try:
            self.subtype = YOLOSubtype(subtype.lower())
        except ValueError as err:
            raise ValueError(
                f"Invalid YOLO subtype {subtype}. Supported YOLO subtypes are {[e.value for e in YOLOSubtype][:-1]}."
            ) from err

        self._logger.debug(f"Subtype set to {self.subtype}")

    def setLabelNames(self, label_names: List[str]) -> None:
        """Sets the names of the classes.

        @param label_names: The names of the classes.
        @type label_names: List[str]
        """
        if not isinstance(label_names, list):
            raise ValueError("Label names must be a list.")
        if not all(isinstance(label, str) for label in label_names):
            raise ValueError("Label names must be a list of strings.")

        self.label_names = label_names
        self._logger.debug(f"Label names set to {self.label_names}")

    def setKeypointLabelNames(self, keypoint_label_names: List[str]) -> None:
        """Sets the label names for the keypoints.

        @param keypoint_label_names: The labels for the keypoints.
        @type keypoint_label_names: List[str]
        """
        if not isinstance(keypoint_label_names, list):
            raise ValueError("Keypoint labels must be a list.")
        if not all(isinstance(label, str) for label in keypoint_label_names):
            raise ValueError("Keypoint labels must be a list of strings.")

        self.keypoint_label_names = keypoint_label_names
        self._logger.debug(f"Keypoint label names set to {self.keypoint_label_names}")

    def setKeypointEdges(self, keypoint_edges: List[Tuple[int, int]]) -> None:
        """Sets the edges for the keypoints.

        @param keypoint_edges: The edges for the keypoints.
        @type keypoint_edges: List[Tuple[int, int]]
        """
        if not isinstance(keypoint_edges, list) and not isinstance(
            keypoint_edges, tuple
        ):
            raise ValueError("Keypoint edges must be a list or tuple.")
        if not all(
            isinstance(edge, tuple)
            and len(edge) == 2
            and all(isinstance(i, int) for i in edge)
            for edge in keypoint_edges
        ):
            raise ValueError("Keypoint edges must be a list of tuples of integers.")

        self.keypoint_edges = keypoint_edges
        self._logger.debug(f"Keypoint edges set to {self.keypoint_edges}")

    def build(
        self,
        head_config: Dict[str, Any],
    ):
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: YOLOExtendedParser
        """

        output_layers = head_config.get("outputs", [])
        bbox_layer_names = [name for name in output_layers if "_yolo" in name]
        kps_layer_names = [name for name in output_layers if "kpt_output" in name]
        masks_layer_names = [name for name in output_layers if "_masks" in name]

        if not bbox_layer_names:
            raise ValueError(
                "YOLOExtendedParser requires the output layers of the bounding boxes detection head."
            )
        if (kps_layer_names and masks_layer_names) or (
            not bbox_layer_names and (kps_layer_names or masks_layer_names)
        ):
            raise ValueError(
                "YOLOExtendedParser requires either the output layers of the keypoints detection head or the output layers of the masks detection head along with the output layers of the bounding boxes detection head but not both."
            )

        self.output_layer_names = bbox_layer_names + kps_layer_names + masks_layer_names

        self.conf_threshold = head_config.get("conf_threshold", self.conf_threshold)
        self.n_classes = head_config.get("n_classes", self.n_classes)
        self.iou_threshold = head_config.get("iou_threshold", self.iou_threshold)
        self.mask_conf = head_config.get("mask_conf", self.mask_conf)
        self.anchors = head_config.get("anchors", self.anchors)
        self.n_keypoints = head_config.get("n_keypoints", self.n_keypoints)
        subtype = head_config.get("subtype", self.subtype)
        self.label_names = head_config.get("classes", self.label_names)
        self.keypoint_label_names = head_config.get(
            "keypoint_label_names", self.keypoint_label_names
        )
        keypoint_edges = head_config.get("skeleton_edges", self.keypoint_edges)
        if keypoint_edges:
            self.keypoint_edges = [tuple(edge) for edge in keypoint_edges]
        try:
            self.subtype = YOLOSubtype(subtype.lower())
        except ValueError as err:
            raise ValueError(
                f"Invalid YOLO subtype {subtype}. Supported YOLO subtypes are {[e.value for e in YOLOSubtype][:-1]}."
            ) from err

        self._logger.debug(
            f"YOLOExtendedParser built with conf_threshold={self.conf_threshold}, n_classes={self.n_classes}, label_names={self.label_names}, iou_threshold={self.iou_threshold}, mask_conf={self.mask_conf}, n_keypoints={self.n_keypoints}, anchors={self.anchors}, subtype='{self.subtype}', keypoint_label_names={self.keypoint_label_names}, keypoint_edges={self.keypoint_edges}"
        )

        return self

    def run(self):
        self._logger.debug("YOLOExtendedParser run started")
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped, no more data
            # Get all the layer names
            layer_names = self.output_layer_names or output.getAllLayerNames()
            self._logger.debug(f"Processing input with layers: {layer_names}")

            outputs_names = sorted(
                [name for name in layer_names if "_yolo" in name or "yolo-" in name]
            )
            outputs_values = [
                output.getTensor(
                    o, dequantize=True, storageOrder=dai.TensorInfo.StorageOrder.NCHW
                ).astype(np.float32)
                for o in outputs_names
            ]

            if (
                any("kpt_output" in name for name in layer_names)
                and self.subtype != YOLOSubtype.P
            ):
                mode = self._KPTS_MODE
                # Get the keypoint outputs
                kpts_output_names = sorted(
                    [name for name in layer_names if "kpt_output" in name]
                )
                kpts_outputs = [
                    output.getTensor(o, dequantize=True).astype(np.float32)
                    for o in kpts_output_names
                ]
            elif (
                any("_masks" in name for name in layer_names)
                and self.subtype != YOLOSubtype.P
            ):
                mode = self._SEG_MODE
                # Get the segmentation outputs
                (
                    masks_outputs_values,
                    protos_output,
                    protos_len,
                ) = get_segmentation_outputs(output)
            else:
                mode = self._DET_MODE

            # Get the model's input shape
            strides = (
                [8, 16, 32]
                if self.subtype
                not in [YOLOSubtype.V3UT, YOLOSubtype.V3T, YOLOSubtype.V4T]
                else [16, 32]
            )
            input_shape = tuple(
                dim * strides[0] for dim in outputs_values[0].shape[2:4]
            )

            # Reshape the anchors based on the model's output heads
            if self.anchors is not None:
                self.anchors = np.array(self.anchors).reshape(len(strides), -1)

            # Ensure the number of classes is correct
            num_classes_check = (
                outputs_values[0].shape[1] - 5
                if self.anchors is None
                else (outputs_values[0].shape[1] // self.anchors.shape[0]) - 5
            )
            if num_classes_check != self.n_classes:
                raise ValueError(
                    f"The provided number of classes {self.n_classes} does not match the model's {num_classes_check}."
                )

            # Ensure the number of keypoints is correct
            if mode == self._KPTS_MODE:
                num_keypoints_check = kpts_outputs[0].shape[1] // 3
                if num_keypoints_check != self.n_keypoints:
                    raise ValueError(
                        f"The provided number of keypoints {self.n_keypoints} does not match the model's {num_keypoints_check}."
                    )

            # Decode the outputs
            results = decode_yolo_output(
                outputs_values,
                strides,
                self.anchors,
                kpts=kpts_outputs if mode == self._KPTS_MODE else None,
                conf_thres=self.conf_threshold,
                iou_thres=self.iou_threshold,
                num_classes=self.n_classes,
                det_mode=mode == self._DET_MODE,
                subtype=self.subtype,
            )

            bboxes, labels, label_names, scores, additional_output = [], [], [], [], []
            final_mask = np.full(input_shape, -1, dtype=np.int16)
            for i in range(results.shape[0]):
                bbox, conf, label, other = (
                    results[i, :4],
                    results[i, 4],
                    results[i, 5].astype(int),
                    results[i, 6:],
                )

                bbox = xyxy_to_xywh(bbox.reshape(1, 4))
                bbox = normalize_bboxes(
                    bbox, height=input_shape[0], width=input_shape[1]
                )[0]
                bboxes.append(bbox)
                labels.append(int(label))
                if self.label_names:
                    label_names.append(self.label_names[int(label)])
                scores.append(conf)

                if mode == self._KPTS_MODE:
                    kpts = parse_kpts(other, self.n_keypoints, input_shape)
                    additional_output.append(kpts)
                elif mode == self._SEG_MODE:
                    seg_coeff = other.astype(int)
                    hi, ai, xi, yi = seg_coeff
                    mask_coeff = masks_outputs_values[hi][
                        0, ai * protos_len : (ai + 1) * protos_len, yi, xi
                    ]
                    mask = process_single_mask(
                        protos_output[0], mask_coeff, self.mask_conf, bbox
                    )
                    # Resize mask to input shape
                    resized_mask = cv2.resize(
                        mask,
                        (input_shape[1], input_shape[0]),
                        interpolation=cv2.INTER_NEAREST,
                    )
                    # Fill the final mask with the instance values
                    final_mask[resized_mask > 0] = i

            bboxes = np.array(bboxes)
            bboxes = np.clip(bboxes, 0, 1)

            if mode == self._KPTS_MODE:
                additional_output = np.array(additional_output)
                keypoints = np.array([])
                keypoints_scores = np.array([])
                if additional_output.size > 0:
                    keypoints = additional_output[:, :, :2]
                    keypoints_scores = additional_output[:, :, 2]

                keypoints = np.clip(keypoints, 0, 1)
                detections_message = create_detection_message(
                    bboxes=bboxes,
                    scores=np.array(scores),
                    labels=np.array(labels),
                    label_names=label_names,
                    keypoints=keypoints,
                    keypoints_scores=keypoints_scores,
                    keypoint_label_names=self.keypoint_label_names,
                    keypoint_edges=self.keypoint_edges,
                )
            elif mode == self._SEG_MODE:
                detections_message = create_detection_message(
                    bboxes=bboxes,
                    scores=np.array(scores),
                    labels=np.array(labels),
                    label_names=label_names,
                    masks=final_mask,
                )
            else:
                detections_message = create_detection_message(
                    bboxes=bboxes,
                    scores=np.array(scores),
                    labels=np.array(labels),
                    label_names=label_names,
                )

            detections_message.setTimestamp(output.getTimestamp())
            detections_message.setTransformation(output.getTransformation())
            detections_message.setSequenceNum(output.getSequenceNum())

            self._logger.debug(
                f"Created detection message with {len(bboxes)} detections"
            )

            self.out.send(detections_message)

            self._logger.debug("Detection message sent successfully")



import numpy as np
import pytest

from depthai_nodes import Line, Lines
from depthai_nodes.message.creators import create_line_detection_message

LINE = np.array([[0.1, 0.2, 0.3, 0.4]])
SCORE = np.array([0.9])


def test_valid_input():
    lines = np.array([[0.1, 0.2, 0.3, 0.4], [0.5, 0.6, 0.7, 0.8]])
    scores = np.array([0.9, 0.8])
    message = create_line_detection_message(lines, scores)

    assert isinstance(message, Lines)
    assert len(message.lines) == 2

    for i, line in enumerate(message.lines):
        assert isinstance(line, Line)
        assert np.allclose(line.start_point.x, lines[i][0], atol=1e-3)
        assert np.allclose(line.start_point.y, lines[i][1], atol=1e-3)
        assert np.allclose(line.end_point.x, lines[i][2], atol=1e-3)
        assert np.allclose(line.end_point.y, lines[i][3], atol=1e-3)
        assert np.allclose(line.confidence, scores[i], atol=1e-3)


def test_empty_lines():
    lines = np.array([])
    scores = np.array([])
    message = create_line_detection_message(lines, scores)

    assert isinstance(message, Lines)
    assert len(message.lines) == 0


def test_invalid_lines_type():
    with pytest.raises(ValueError):
        create_line_detection_message(LINE.tolist(), SCORE)


def test_invalid_lines_shape():
    with pytest.raises(ValueError):
        create_line_detection_message(LINE[0], SCORE)


def test_invalid_lines_dimension():
    with pytest.raises(ValueError):
        create_line_detection_message(LINE[:, :3], SCORE)


def test_invalid_scores_type():
    with pytest.raises(ValueError):
        create_line_detection_message(LINE, SCORE.tolist())


def test_invalid_scores_shape():
    with pytest.raises(ValueError):
        create_line_detection_message(LINE, np.array([SCORE]))


def test_invalid_scores_value_type():
    with pytest.raises(ValueError):
        create_line_detection_message(LINE, np.array([1], dtype=np.int64))


def test_mismatched_lines_scores_length():
    with pytest.raises(ValueError):
        create_line_detection_message(LINE, np.array([0.9, 0.8]))



from typing import List, Tuple, Type

import depthai as dai

from .device import DeviceMock
from .input import InfiniteInputMock, InputMock
from .output import OutputMock
from .threaded_host_node import ThreadedHostNodeMock


class PipelineMock:
    """Mock class for the depthai Pipeline.

    The class is used to create a mock pipeline for testing purposes.
    """

    def __init__(self):
        self._nodes = []
        self._defaultDevice = DeviceMock()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def getDefaultDevice(self):
        return self._defaultDevice

    def remove(self, node):
        self._nodes.remove(node)

    def create(self, node_type: Type[ThreadedHostNodeMock]):
        from depthai_nodes.node.parsers.base_parser import BaseParser

        if issubclass(node_type, BaseParser):
            # Create a concrete subclass of the parser that implements the abstract methods
            class ParserMock(node_type):
                """Concrete parser class that implements the abstract methods of the
                BaseParser."""

                def __init__(self, parent_pipeline: PipelineMock):
                    self.parent_pipeline = parent_pipeline
                    super().__init__()
                    self._input = InfiniteInputMock()
                    self._out = OutputMock()
                    self._is_running = True

                @property
                def input(self):
                    return self._input

                @input.setter
                def input(self, input: InfiniteInputMock):
                    self._input = input

                @property
                def out(self):
                    return self._out

                @out.setter
                def out(self, output: OutputMock):
                    self._out = output

                def createInput(self):
                    self._input = InfiniteInputMock()
                    return self._input

                def createOutput(
                    self, possibleDatatypes: List[Tuple[dai.DatatypeEnum, bool]] = None
                ):
                    self._out = OutputMock()
                    return self._out

                def process(self, data: dai.NNData):
                    pass

                def isRunning(self):
                    return self._is_running

                def setIsRunning(self, is_running: bool):
                    self._is_running = is_running

                def getParentPipeline(self):
                    return self.parent_pipeline

            node = ParserMock(self)

        elif node_type == dai.node.DetectionParser:
            node = DetectionParserMock()

        else:

            class NodeMock(node_type):
                def __init__(self, pipeline):
                    self._pipeline = pipeline
                    super().__init__()
                    if (
                        self.__class__.__bases__[0].__bases__[0].__name__
                        == "ThreadedHostNodeMock"
                    ):
                        self._input = InfiniteInputMock()
                    else:
                        self._input = InputMock()
                    self._out = OutputMock()
                    self._is_running = True

                def getParentPipeline(self):
                    return self._pipeline

                @property
                def input(self):
                    return self._input

                @input.setter
                def input(self, input):
                    self._input = input

                @property
                def out(self):
                    return self._out

                @out.setter
                def out(self, output):
                    self._out = output

                def setNNArchive(self, nn_archive):
                    # check if superclass has the method
                    if hasattr(super(), "setNNArchive"):
                        super().setNNArchive(nn_archive)
                    else:
                        self._nn_archive = nn_archive

                def createInput(self):
                    if (
                        self.__class__.__bases__[0].__bases__[0].__name__
                        == "ThreadedHostNodeMock"
                    ):
                        self._input = InfiniteInputMock()
                    else:
                        self._input = InputMock()
                    return self._input

                def isRunning(self):
                    return self._is_running

                def setIsRunning(self, is_running: bool):
                    self._is_running = is_running

                def createOutput(
                    self, possibleDatatypes: List[Tuple[dai.DatatypeEnum, bool]] = None
                ):
                    self._out = OutputMock()
                    return self._out

            node = NodeMock(self)

        node.parentPipeline = self
        self._nodes.append(node)
        return node


class DetectionParserMock:
    def __init__(self):
        self._input = InputMock()
        self._out = OutputMock()
        self._nn_archive = None

    @property
    def input(self):
        return self._input

    @input.setter
    def input(self, input):
        self._input = input

    @property
    def out(self):
        return self._out

    @out.setter
    def out(self, output):
        self._out = output

    def build(self):
        pass

    def setNNArchive(self, nn_archive):
        self._nn_archive = nn_archive



from .input import InputMock
from .node import NodeMock


class AutoCreateDict(dict):
    """Dictionary that automatically creates entries when accessed."""

    def __getitem__(self, key):
        if key not in self:
            self[key] = InputMock()
        return super().__getitem__(key)


class SyncMock(NodeMock):
    def __init__(self):
        super().__init__()
        self.inputs = AutoCreateDict()

    def setRunOnHost(self, run_on_host: bool):
        self._run_on_host = run_on_host

    def runOnHost(self):
        return self._run_on_host



from typing import Any, Dict

import depthai as dai
import numpy as np

from depthai_nodes.message.creators import create_segmentation_message
from depthai_nodes.node.parsers.base_parser import BaseParser


class SegmentationParser(BaseParser):
    """Parser class for parsing the output of the segmentation models.

    Attributes
    ----------
    output_layer_name: str
        Name of the output layer relevant to the parser.
    classes_in_one_layer : bool
        Whether all classes are in one layer in the multi-class segmentation model. Default is False. If True, the parser will use np.max instead of np.argmax to get the class map.

    Output Message/s
    ----------------
    **Type**: SegmentationMask

    **Description**: Segmentation message containing the segmentation mask. Every pixel belongs to exactly one class. Unassigned pixels are represented with "-1" and class pixels with non-negative integers.

    Error Handling
    --------------
    **ValueError**: If the number of output layers is not E{1}.

    **ValueError**: If the number of dimensions of the output tensor is not E{3}.
    """

    def __init__(
        self, output_layer_name: str = "", classes_in_one_layer: bool = False
    ) -> None:
        """Initializes the parser node.

        @param output_layer_name: Name of the output layer relevant to the parser.
        @type output_layer_name: str
        @param classes_in_one_layer: Whether all classes are in one layer in the multi-
            class segmentation model. Default is False. If True, the parser will use
            np.max instead of np.argmax to get the class map.
        @type classes_in_one_layer: bool
        """
        super().__init__()
        self.output_layer_name = output_layer_name
        self.classes_in_one_layer = classes_in_one_layer
        self._logger.debug(
            f"SegmentationParser initialized with output_layer_name='{output_layer_name}', classes_in_one_layer={classes_in_one_layer}"
        )

    def setOutputLayerName(self, output_layer_name: str) -> None:
        """Sets the name of the output layer.

        @param output_layer_name: The name of the output layer.
        @type output_layer_name: str
        """
        if not isinstance(output_layer_name, str):
            raise ValueError("Output layer name must be a string.")
        self.output_layer_name = output_layer_name
        self._logger.debug(f"Output layer name set to '{self.output_layer_name}'")

    def setClassesInOneLayer(self, classes_in_one_layer: bool) -> None:
        """Sets the flag indicating whether all classes are in one layer.

        @param classes_in_one_layer: Whether all classes are in one layer.
        @type classes_in_one_layer: bool
        """
        if not isinstance(classes_in_one_layer, bool):
            raise ValueError("classes_in_one_layer must be a boolean.")
        self.classes_in_one_layer = classes_in_one_layer
        self._logger.debug(f"Classes in one layer set to {self.classes_in_one_layer}")

    def build(
        self,
        head_config: Dict[str, Any],
    ) -> "SegmentationParser":
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: SegmentationParser
        """

        output_layers = head_config.get("outputs", [])
        if len(output_layers) != 1:
            raise ValueError(
                f"Only one output layer supported for Segmentation, got {output_layers} layers."
            )
        self.output_layer_name = output_layers[0]
        self.classes_in_one_layer = head_config.get(
            "classes_in_one_layer", self.classes_in_one_layer
        )

        self._logger.debug(
            f"SegmentationParser built with output_layer_name='{self.output_layer_name}', classes_in_one_layer={self.classes_in_one_layer}"
        )

        return self

    def run(self):
        self._logger.debug("SegmentationParser run started")
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            layers = output.getAllLayerNames()
            self._logger.debug(f"Processing input with layers: {layers}")
            if len(layers) == 1 and self.output_layer_name == "":
                self.output_layer_name = layers[0]
            elif len(layers) != 1 and self.output_layer_name == "":
                raise ValueError(
                    f"Expected 1 output layer, got {len(layers)} layers. Please provide the output_layer_name."
                )

            segmentation_mask = output.getTensor(
                self.output_layer_name, dequantize=True
            )
            if len(segmentation_mask.shape) == 4:
                segmentation_mask = segmentation_mask[0]

            if len(segmentation_mask.shape) != 3:
                raise ValueError(
                    f"Expected 3D output tensor, got {len(segmentation_mask.shape)}D."
                )

            np_function = np.argmax
            mask_shape = segmentation_mask.shape
            min_dim = np.argmin(mask_shape)
            if min_dim == len(mask_shape) - 1:
                segmentation_mask = segmentation_mask.transpose(2, 0, 1)
            adding_unassigned_class = False
            if segmentation_mask.shape[0] == 1:  # shape is (1, H, W)
                if self.classes_in_one_layer:
                    np_function = np.max
                else:
                    # If there is only one class, add an unassigned class
                    adding_unassigned_class = True
                    segmentation_mask = np.vstack(
                        (
                            np.zeros(
                                (
                                    1,
                                    segmentation_mask.shape[1],
                                    segmentation_mask.shape[2],
                                ),
                                dtype=np.float32,
                            ),
                            segmentation_mask,
                        )
                    )

            class_map = (
                np_function(segmentation_mask, axis=0)
                .reshape(segmentation_mask.shape[1], segmentation_mask.shape[2])
                .astype(np.int16)
            )

            if adding_unassigned_class:
                class_map = class_map - 1

            mask_message = create_segmentation_message(class_map)
            mask_message.setTimestamp(output.getTimestamp())
            mask_message.setTransformation(output.getTransformation())
            mask_message.setSequenceNum(output.getSequenceNum())

            self._logger.debug(
                f"Created segmentation message with {class_map.shape[0]} classes"
            )

            self.out.send(mask_message)

            self._logger.debug("Segmentation message sent successfully")



import depthai as dai
import numpy as np
import pytest

from depthai_nodes.message.creators import (
    create_tracked_features_message,
)

REFERENCE_POINTS = np.array([[0.1, 0.2], [0.3, 0.4]])
TARGET_POINTS = np.array([[0.5, 0.6], [0.7, 0.8]])


def test_valid_input():
    message = create_tracked_features_message(REFERENCE_POINTS, TARGET_POINTS)

    assert isinstance(message, dai.TrackedFeatures)
    assert len(message.trackedFeatures) == 4
    assert all(
        isinstance(feature, dai.TrackedFeature) for feature in message.trackedFeatures
    )

    expected_values = [
        (0.1, 0.2, 0, 0),
        (0.5, 0.6, 0, 1),
        (0.3, 0.4, 1, 0),
        (0.7, 0.8, 1, 1),
    ]

    for feature, (expected_x, expected_y, expected_id, expected_age) in zip(
        message.trackedFeatures, expected_values
    ):
        assert np.allclose(feature.position.x, expected_x, atol=1e-3)
        assert np.allclose(feature.position.y, expected_y, atol=1e-3)
        assert np.allclose(feature.id, expected_id, atol=1e-3)
        assert np.allclose(feature.age, expected_age, atol=1e-3)


def test_invalid_reference_points_type():
    with pytest.raises(ValueError):
        create_tracked_features_message(REFERENCE_POINTS.tolist(), TARGET_POINTS)


def test_invalid_reference_points_shape():
    with pytest.raises(ValueError):
        create_tracked_features_message(REFERENCE_POINTS.flatten(), TARGET_POINTS)


def test_invalid_reference_points_dimension():
    reference_points = np.array([[0.1, 0.2, 0.3], [0.4, 0.5, 0.6]])
    with pytest.raises(ValueError):
        create_tracked_features_message(reference_points, TARGET_POINTS)


def test_invalid_target_points_type():
    with pytest.raises(ValueError):
        create_tracked_features_message(REFERENCE_POINTS, TARGET_POINTS.tolist())


def test_invalid_target_points_shape():
    with pytest.raises(ValueError):
        create_tracked_features_message(REFERENCE_POINTS, TARGET_POINTS.flatten())


def test_invalid_target_points_dimension():
    target_points = np.array([[0.5, 0.6, 0.7], [0.8, 0.9, 1.0]])
    with pytest.raises(ValueError):
        create_tracked_features_message(REFERENCE_POINTS, target_points)


def test_mismatched_points_length():
    target_points = np.array([[0.5, 0.6]])
    with pytest.raises(ValueError):
        create_tracked_features_message(REFERENCE_POINTS, target_points)



import copy
from typing import Union

import depthai as dai


def copy_message(msg: dai.Buffer) -> dai.Buffer:
    """Copies the incoming message and returns it.

    @param msg: The input message.
    @type msg: dai.Buffer
    @return: The copied message.
    @rtype: dai.Buffer
    """

    # 1st attempt: native .copy() method
    if hasattr(msg, "copy"):
        return msg.copy()

    # 2nd attempt: custom copy implementation
    try:
        return _copy(msg)
    except TypeError:
        pass

    # 3rd attempt: deepcopy (the most general approach)
    try:
        return copy.deepcopy(msg)
    except TypeError as e:
        raise TypeError(f"Copying of message type {type(msg)} is not supported.") from e


def _copy(msg: dai.Buffer) -> dai.Buffer:
    def _copy_metadata(msg: dai.Buffer) -> dai.Buffer:
        msg_type = type(msg)
        msg_copy = msg_type()
        if hasattr(msg, "getSequenceNum"):
            msg_copy.setSequenceNum(msg.getSequenceNum())
        if hasattr(msg, "getTimestamp"):
            msg_copy.setTimestamp(msg.getTimestamp())
        if hasattr(msg, "getTimestampDevice"):
            msg_copy.setTimestampDevice(msg.getTimestampDevice())
        if hasattr(msg, "getTransformation"):
            msg_copy.setTransformation(msg.getTransformation())
        return msg_copy

    def _copy_img_frame(img_frame: dai.ImgFrame) -> dai.ImgFrame:
        img_frame_copy = _copy_metadata(img_frame)
        img_frame_copy.setCvFrame(img_frame.getCvFrame(), img_frame.getType())
        img_frame_copy.setCategory(img_frame.getCategory())
        return img_frame_copy

    def _copy_img_detection(
        img_det: Union[dai.ImgDetection, dai.SpatialImgDetection],
    ) -> Union[dai.ImgDetection, dai.SpatialImgDetection]:
        assert isinstance(img_det, (dai.ImgDetection, dai.SpatialImgDetection))
        img_det_copy = _copy_metadata(img_det)
        if isinstance(img_det, dai.SpatialImgDetection):
            img_det_copy.spatialCoordinates = img_det.spatialCoordinates
            img_det_copy.boundingBoxMapping = img_det.boundingBoxMapping
        img_det_copy.xmin = img_det.xmin
        img_det_copy.ymin = img_det.ymin
        img_det_copy.xmax = img_det.xmax
        img_det_copy.ymax = img_det.ymax
        img_det_copy.label = img_det.label
        img_det_copy.confidence = img_det.confidence
        return img_det_copy

    def _copy_img_detections(
        img_dets: Union[dai.ImgDetections, dai.SpatialImgDetections],
    ) -> Union[dai.ImgDetections, dai.SpatialImgDetections]:
        assert isinstance(img_dets, (dai.ImgDetections, dai.SpatialImgDetections))
        img_dets_copy = _copy_metadata(img_dets)
        img_dets_copy.detections = [
            _copy_img_detection(img_det) for img_det in img_dets.detections
        ]
        return img_dets_copy

    def _copy_point2f(point2f: dai.Point2f) -> dai.Point2f:
        point2f_copy = _copy_metadata(point2f)
        point2f_copy.x = point2f.x
        point2f_copy.y = point2f.y
        # TODO: set the value for .isNormalized()
        return point2f_copy

    if isinstance(msg, dai.ImgFrame):
        return _copy_img_frame(msg)
    elif isinstance(msg, (dai.ImgDetection, dai.SpatialImgDetection)):
        return _copy_img_detection(msg)
    elif isinstance(msg, (dai.ImgDetections, dai.SpatialImgDetections)):
        return _copy_img_detections(msg)
    elif isinstance(msg, dai.Point2f):
        return _copy_point2f(msg)
    else:
        # TODO: define logic for copying other message types
        raise TypeError(f"Copying of message type {type(msg)} is not supported.")



import depthai as dai

devices = dai.Device.getAllAvailableDevices()
RVC2_IP = ""
RVC4_IP = ""
for device in devices:
    mxid = device.getDeviceId()
    platform = device.platform.name
    if "RVC4" in platform:
        RVC4_IP = str(mxid)
    if "MYRIAD" in platform:
        RVC2_IP = str(mxid)

# set env variables
print(f"export RVC2_IP={RVC2_IP}")
print(f"export RVC4_IP={RVC4_IP}")



from typing import List

import depthai as dai

from depthai_nodes import ImgDetectionsExtended
from depthai_nodes.message.utils import copy_message
from depthai_nodes.node.base_host_node import BaseHostNode


class ImgDetectionsFilter(BaseHostNode):
    """Filters out detections based on the specified criteria and outputs them as a separate message.
    The order of operations:
        1. Filter by label/confidence;
        2. Sort (if applicable);
        3. Subset.

    Attributes
    ----------
    labels_to_keep : List[int]
        Labels to keep. Only detections with labels in this list will be kept.
    labels_to_reject: List[int]
        Labels to reject. Only detections with labels not in this list will be kept.
    confidence_threshold : float
        Minimum confidence threshold. Detections with confidence below this threshold will be filtered out.
    max_detections : int
        Maximum number of detections to keep. If not defined, all detections will be kept.
    sort_by_confidence: bool
        Whether to sort the detections by confidence before subsetting. If True, the detections will be sorted
            in descending order of confidence. It's set to False by default.
    """

    def __init__(self):
        super().__init__()
        self._labels_to_keep = None
        self._labels_to_reject = None
        self._confidence_threshold = None
        self._max_detections = None
        self._sort_by_confidence = False
        self._logger.debug("ImgDetectionsFilter initialized")

    def setLabels(self, labels: List[int], keep: bool) -> None:
        """Sets the labels to keep or reject.

        @param labels: The labels to keep or reject.
        @type labels: List[int]
        @param keep: Whether to keep or reject the labels.
        @type keep: bool
        """
        if not isinstance(labels, list) and labels is not None:
            raise ValueError("Labels must be a list or None.")
        if not isinstance(keep, bool):
            raise ValueError("keep must be a boolean.")

        if keep:
            self._labels_to_keep = labels
            self._labels_to_reject = None
        else:
            self._labels_to_keep = None
            self._labels_to_reject = labels

        self._logger.debug(
            f"Labels set to {self._labels_to_keep if keep else self._labels_to_reject}"
        )

    def setConfidenceThreshold(self, confidence_threshold: float) -> None:
        """Sets the confidence threshold.

        @param confidence_threshold: The confidence threshold.
        @type confidence_threshold: float
        """
        if (
            not isinstance(confidence_threshold, float)
            and confidence_threshold is not None
        ):
            raise ValueError("confidence_threshold must be a float.")
        self._confidence_threshold = confidence_threshold
        self._logger.debug(f"Confidence threshold set to {self._confidence_threshold}")

    def setMaxDetections(self, max_detections: int) -> None:
        """Sets the maximum number of detections.

        @param max_detections: The maximum number of detections.
        @type max_detections: int
        """
        if not isinstance(max_detections, int) and max_detections is not None:
            raise ValueError("max_detections must be an integer.")
        self._max_detections = max_detections
        self._logger.debug(f"Max detections set to {self._max_detections}")

    def setSortByConfidence(self, sort_by_confidence: bool) -> None:
        """Sets whether to sort the detections by confidence before subsetting.

        @param sort_by_confidence: Whether to sort the detections.
        @type sort_by_confidence: bool
        """
        if not isinstance(sort_by_confidence, bool):
            raise ValueError("sort_by_confidence must be a boolean.")
        self._sort_by_confidence = sort_by_confidence
        self._logger.debug(f"Sort by confidence set to {self._sort_by_confidence}")

    def build(
        self,
        msg: dai.Node.Output,
        labels_to_keep: List[int] = None,
        labels_to_reject: List[int] = None,
        confidence_threshold: float = None,
        max_detections: int = None,
        sort_by_confidence: bool = False,
    ) -> "ImgDetectionsFilter":
        self.link_args(msg)

        if labels_to_keep is not None:
            if labels_to_reject is not None:
                raise ValueError(
                    "Cannot set labels_to_keep and labels_to_reject at the same time."
                )
            else:
                self.setLabels(labels_to_keep, keep=True)
        else:
            if labels_to_reject is not None:
                self.setLabels(labels_to_reject, keep=False)

        if confidence_threshold is not None:
            self.setConfidenceThreshold(confidence_threshold)

        if max_detections is not None:
            self.setMaxDetections(max_detections)

        self.setSortByConfidence(sort_by_confidence)

        self._logger.debug(
            f"ImgDetectionsFilter built with labels_to_keep={labels_to_keep}, labels_to_reject={labels_to_reject}, confidence_threshold={confidence_threshold}, max_detections={max_detections}, sort_by_confidence={sort_by_confidence}"
        )

        return self

    def process(self, msg: dai.Buffer) -> None:
        self._logger.debug("Processing new input")
        assert isinstance(
            msg,
            (
                dai.ImgDetections,
                ImgDetectionsExtended,
                dai.SpatialImgDetections,
            ),
        )

        msg_new = copy_message(
            msg
        )  # we don't want to modify the original message as it might be used by other nodes

        filtered_detections = []
        for detection in msg.detections:
            if self._labels_to_keep is not None:
                if detection.label not in self._labels_to_keep:
                    continue

            if self._labels_to_reject is not None:
                if detection.label in self._labels_to_reject:
                    continue

            if self._confidence_threshold is not None:
                if detection.confidence < self._confidence_threshold:
                    continue

            filtered_detections.append(detection)

        if self._sort_by_confidence:
            filtered_detections = sorted(
                filtered_detections, key=lambda x: x.confidence, reverse=True
            )

        msg_new.detections = filtered_detections[: self._max_detections]

        self._logger.debug("Detections message created")

        self.out.send(msg_new)

        self._logger.debug("Message sent successfully")



import depthai as dai
import numpy as np
import pytest

from depthai_nodes import Line, Lines


@pytest.fixture
def line():
    return Line()


@pytest.fixture
def lines():
    return Lines()


def test_line_initialization(line: Line):
    assert line.start_point is None
    assert line.end_point is None
    assert line.confidence is None


def test_line_set_start_point(line: Line):
    start_point = dai.Point2f(0.1, 0.2)
    line.start_point = start_point
    assert np.allclose(line.start_point.x, 0.1, atol=1e-3)
    assert np.allclose(line.start_point.y, 0.2, atol=1e-3)

    with pytest.raises(TypeError):
        line.start_point = "not a Point2f"


def test_line_set_end_point(line: Line):
    end_point = dai.Point2f(0.3, 0.4)
    line.end_point = end_point
    assert np.allclose(line.end_point.x, 0.3, atol=1e-3)
    assert np.allclose(line.end_point.y, 0.4, atol=1e-3)

    with pytest.raises(TypeError):
        line.end_point = "not a Point2f"


def test_line_set_confidence(line: Line):
    line.confidence = 0.9
    assert line.confidence == 0.9

    with pytest.raises(TypeError):
        line.confidence = "not a float"


def test_lines_initialization(lines: Lines):
    assert lines.lines == []
    assert lines.transformation is None


def test_lines_set_lines(lines: Lines):
    line1 = Line()
    line2 = Line()
    lines_list = [line1, line2]
    lines.lines = lines_list
    assert lines.lines == lines_list

    with pytest.raises(TypeError):
        lines.lines = "not a list"

    with pytest.raises(ValueError):
        lines.lines = [line1, "not a Line"]


def test_lines_set_transformation(lines: Lines):
    transformation = dai.ImgTransformation()
    lines.transformation = transformation
    assert lines.transformation == transformation

    with pytest.raises(TypeError):
        lines.transformation = "not a dai.ImgTransformation"


def test_lines_set_transformation_none(lines: Lines):
    lines.transformation = None
    assert lines.transformation is None



import numpy as np

from depthai_nodes import SegmentationMask


def create_segmentation_message(mask: np.ndarray) -> SegmentationMask:
    """Create a DepthAI message for segmentation mask.

    @param mask: Segmentation map array of shape (H, W) where each value represents a
        segmented object class.
    @type mask: np.array
    @return: Segmentation mask message.
    @rtype: SegmentationMask
    @raise ValueError: If mask is not a numpy array.
    @raise ValueError: If mask is not 2D.
    @raise ValueError: If mask is not of type int16.
    """

    if not isinstance(mask, np.ndarray):
        raise ValueError(f"Expected numpy array, got {type(mask)}.")

    if len(mask.shape) != 2:
        raise ValueError(f"Expected 2D input, got {len(mask.shape)}D input.")

    if mask.dtype != np.int16:
        raise ValueError(f"Expected int16 input, got {mask.dtype}.")

    mask_msg = SegmentationMask()
    mask_msg.mask = mask

    return mask_msg



from typing import Any, Dict, List, Optional, Tuple

import cv2
import numpy as np

from depthai_nodes.node.parsers.utils import sigmoid
from depthai_nodes.node.parsers.utils.yolo import (
    non_max_suppression,
    parse_yolo_outputs,
)


def box_prompt(
    masks: np.ndarray, bbox: Tuple[int, int, int, int], orig_shape: Tuple[int, int]
) -> np.ndarray:
    """Modifies the bounding box properties and calculates IoU between masks and
    bounding box.

    Source: https://github.com/ultralytics/ultralytics/blob/main/ultralytics/models/fastsam/prompt.py#L286
    Modified so it uses numpy instead of torch.

    @param masks: The resulting masks of the FastSAM model
    @type masks: np.ndarray
    @param bbox: The prompt bounding box coordinates
    @type bbox: Tuple[int, int, int, int]
    @param orig_shape: The original shape of the image
    @type orig_shape: Tuple[int, int] (height, width)
    @return: The modified masks
    @rtype: np.ndarray
    """
    if masks is not None:
        assert bbox[2] != 0 and bbox[3] != 0
        target_height, target_width = orig_shape
        h, w = masks.shape[1:3]
        if h != target_height or w != target_width:
            bbox = [
                int(bbox[0] * w / target_width),
                int(bbox[1] * h / target_height),
                int(bbox[2] * w / target_width),
                int(bbox[3] * h / target_height),
            ]
        bbox[0] = max(round(bbox[0]), 0)
        bbox[1] = max(round(bbox[1]), 0)
        bbox[2] = min(round(bbox[2]), w)
        bbox[3] = min(round(bbox[3]), h)

        bbox_area = (bbox[3] - bbox[1]) * (bbox[2] - bbox[0])

        masks_area = np.sum(masks[:, bbox[1] : bbox[3], bbox[0] : bbox[2]], axis=(1, 2))
        orig_masks_area = np.sum(masks, axis=(1, 2))

        union = bbox_area + orig_masks_area - masks_area
        iou = masks_area / union
        max_iou_index = np.argmax(iou)

        masks = masks[max_iou_index]
    return masks.reshape(1, masks.shape[0], masks.shape[1])


def format_results(
    bboxes: np.ndarray, masks: np.ndarray, filter: int = 0
) -> List[Dict[str, Any]]:
    """Formats detection results into list of annotations each containing ID,
    segmentation, bounding box, score and area.

    Source: https://github.com/ultralytics/ultralytics/blob/main/ultralytics/models/fastsam/prompt.py#L56

    @param bboxes: The bounding boxes of the detected objects
    @type bboxes: np.ndarray
    @param masks: The masks of the detected objects
    @type masks: np.ndarray
    @param filter: The filter value
    @type filter: int
    @return: The formatted annotations
    @rtype: List[Dict[str, Any]]
    """
    annotations = []
    n = len(masks) if masks is not None else 0
    for i in range(n):
        mask = masks[i] == 1.0
        if np.sum(mask) >= filter:
            annotation = {
                "id": i,
                "segmentation": mask,
                "bbox": bboxes[i][:4],
                "score": bboxes[i][4],
            }
            annotation["area"] = annotation["segmentation"].sum()
            annotations.append(annotation)
    return annotations


def point_prompt(
    bboxes: np.ndarray,
    masks: np.ndarray,
    points: List[Tuple[int, int]],
    pointlabel: List[int],
    orig_shape: Tuple[int, int],
) -> np.ndarray:
    """Adjusts points on detected masks based on user input and returns the modified
    results.

    Source: https://github.com/ultralytics/ultralytics/blob/main/ultralytics/models/fastsam/prompt.py#L321
    Modified so it uses numpy instead of torch.

    @param bboxes: The bounding boxes of the detected objects
    @type bboxes: np.ndarray
    @param masks: The masks of the detected objects
    @type masks: np.ndarray
    @param points: The points to adjust
    @type points: List[Tuple[int, int]]
    @param pointlabel: The point labels
    @type pointlabel: List[int]
    @param orig_shape: The original shape of the image
    @type orig_shape: Tuple[int, int] (height, width)
    @return: The modified masks
    @rtype: np.ndarray
    """
    if masks is not None:
        masks = format_results(bboxes, masks, 0)
        target_height, target_width = orig_shape
        h = masks[0]["segmentation"].shape[0]
        w = masks[0]["segmentation"].shape[1]
        if h != target_height or w != target_width:
            points = [
                [int(point[0] * w / target_width), int(point[1] * h / target_height)]
                for point in points
            ]
        onemask = np.zeros((h, w))
        for annotation in masks:
            mask = (
                annotation["segmentation"]
                if isinstance(annotation, dict)
                else annotation
            )
            for i, point in enumerate(points):
                if mask[point[1], point[0]] == 1 and pointlabel[i] == 1:
                    onemask += mask
                if mask[point[1], point[0]] == 1 and pointlabel[i] == 0:
                    onemask -= mask
        onemask = onemask >= 1
        masks = np.array([onemask])
    return masks


def adjust_bboxes_to_image_border(
    boxes: np.ndarray, image_shape: Tuple[int, int], threshold: int = 20
) -> np.ndarray:
    """
    Source: https://github.com/ultralytics/ultralytics/blob/main/ultralytics/models/fastsam/utils.py#L6 (Ultralytics)
    Adjust bounding boxes to stick to image border if they are within a certain threshold.

    @param boxes: Bounding boxes
    @type boxes: np.ndarray
    @param image_shape: Image shape
    @type image_shape: Tuple[int, int]
    @param threshold: Pixel threshold
    @type threshold: int
    @return: Adjusted bounding boxes
    @rtype: np.ndarray
    """
    # Image dimensions
    h, w = image_shape

    # Adjust boxes
    boxes[boxes[:, 0] < threshold, 0] = 0  # x1
    boxes[boxes[:, 1] < threshold, 1] = 0  # y1
    boxes[boxes[:, 2] > w - threshold, 2] = w  # x2
    boxes[boxes[:, 3] > h - threshold, 3] = h  # y2
    return boxes


def bbox_iou(
    box1: np.ndarray,
    boxes: np.ndarray,
    iou_thres: float = 0.9,
    image_shape: Tuple[int, int] = (640, 640),
    raw_output: bool = False,
) -> np.ndarray:
    """
    Source: https://github.com/ultralytics/ultralytics/blob/main/ultralytics/models/fastsam/utils.py#L30 (Ultralytics - rewritten to numpy)
    Compute the Intersection-Over-Union of a bounding box with respect to an array of other bounding boxes.

    @param box1: Array of shape (4, ) representing a single bounding box.
    @type box1: np.ndarray
    @param boxes: Array of shape (n, 4) representing multiple bounding boxes.
    @type boxes: np.ndarray
    @param iou_thres: IoU threshold
    @type iou_thres: float
    @param image_shape: Image shape (height, width)
    @type image_shape: Tuple[int, int]
    @param raw_output: If True, return the raw IoU values instead of the indices
    @type raw_output: bool
    @return: Indices of boxes with IoU > thres, or the raw IoU values if raw_output is True
    @rtype: np.ndarray
    """
    boxes = adjust_bboxes_to_image_border(boxes, image_shape)

    # Obtain coordinates for intersections
    x1 = np.maximum(box1[0], boxes[:, 0])
    y1 = np.maximum(box1[1], boxes[:, 1])
    x2 = np.minimum(box1[2], boxes[:, 2])
    y2 = np.minimum(box1[3], boxes[:, 3])

    # Compute the area of intersection
    intersection = np.maximum(0, x2 - x1) * np.maximum(0, y2 - y1)

    # Compute the area of both individual boxes
    box1_area = (box1[2] - box1[0]) * (box1[3] - box1[1])
    box2_area = (boxes[:, 2] - boxes[:, 0]) * (boxes[:, 3] - boxes[:, 1])

    # Compute the area of union
    union = box1_area + box2_area - intersection

    # Compute the IoU
    iou = intersection / union  # Should be shape (n, )
    if raw_output:
        return iou if iou.size > 0 else np.array([])

    # Return indices of boxes with IoU > thres
    return np.flatnonzero(iou > iou_thres)


def decode_fastsam_output(
    outputs: List[np.ndarray],
    strides: List[int],
    anchors: List[Optional[np.ndarray]],
    img_shape: Tuple[int, int],
    conf_thres: float = 0.5,
    iou_thres: float = 0.45,
    num_classes: int = 1,
) -> np.ndarray:
    """Decode the output of the FastSAM model.

    @param outputs: List of FastSAM outputs
    @type outputs: List[np.ndarray]
    @param strides: List of strides
    @type strides: List[int]
    @param anchors: List of anchors
    @type anchors: List[Optional[np.ndarray]]
    @param img_shape: Image shape
    @type img_shape: Tuple[int, int]
    @param conf_thres: Confidence threshold
    @type conf_thres: float
    @param iou_thres: IoU threshold
    @type iou_thres: float
    @param num_classes: Number of classes
    @type num_classes: int
    @return: NMS output
    @rtype: np.ndarray
    """
    output = parse_yolo_outputs(
        outputs=outputs, strides=strides, num_outputs=6, anchors=anchors, kpts=None
    )
    output_nms = non_max_suppression(
        output,
        conf_thres=conf_thres,
        iou_thres=iou_thres,
        num_classes=num_classes,
        kpts_mode=False,
    )[0]

    full_box = np.zeros(output_nms.shape[1])
    full_box[2], full_box[3], full_box[4], full_box[6:] = (
        img_shape[1],
        img_shape[0],
        1.0,
        1.0,
    )
    full_box = full_box.reshape((1, -1))
    critical_iou_index = bbox_iou(
        full_box[0][:4], output_nms[:, :4], iou_thres=0.9, image_shape=img_shape
    )

    if critical_iou_index.size > 0:
        full_box[0][4] = output_nms[critical_iou_index][:, 4]
        full_box[0][6:] = output_nms[critical_iou_index][:, 6:]
        output_nms[critical_iou_index] = full_box

    return output_nms


def crop_mask(masks: np.ndarray, box: np.ndarray) -> np.ndarray:
    """It takes a mask and a bounding box, and returns a mask that is cropped to the
    bounding box.

    @param masks: [h, w] array of masks
    @type masks: np.ndarray
    @param box: An array of bbox coordinates in (x1, y1, x2, y2) format
    @type box: np.ndarray
    @return: The masks are being cropped to the bounding box.
    @rtype: np.ndarray
    """
    h, w = masks.shape
    x1, y1, x2, y2 = box
    r = np.arange(w).reshape(1, w)
    c = np.arange(h).reshape(h, 1)
    return masks * ((r >= x1) * (r < x2) * (c >= y1) * (c < y2))


def process_single_mask(
    protos: np.ndarray,
    mask_coeff: np.ndarray,
    mask_conf: float,
    img_shape: Tuple[int, int],
    bbox: Tuple[int, int, int, int],
) -> np.ndarray:
    """Processes a single mask.

    @param protos: Prototypes
    @type protos: np.ndarray
    @param mask_coeff: Mask coefficients
    @type mask_coeff: np.ndarray
    @param mask_conf: Mask confidence
    @type mask_conf: float
    @param img_shape: Image shape
    @type img_shape: Tuple[int, int]
    @param bbox: Bounding box
    @type bbox: Tuple[int, int, int, int]
    @return: Processed mask
    @rtype: np.ndarray
    """
    mask = sigmoid(np.sum(protos * mask_coeff[..., np.newaxis, np.newaxis], axis=0))
    mask = cv2.resize(mask, img_shape, interpolation=cv2.INTER_NEAREST)
    mask = crop_mask(mask, np.array(bbox))
    return (mask > mask_conf).astype(np.uint8)


def merge_masks(masks: np.ndarray) -> np.ndarray:
    """Merge masks to a 2D array where each object is represented by a unique label.

    @param masks: 3D array of masks
    @type masks: np.ndarray
    @return: 2D array of masks
    @rtype: np.ndarray
    """
    if masks.ndim == 3:
        n, height, width = masks.shape
    else:
        raise ValueError("Masks must be a 3D array.")

    merged_masks = np.full((height, width), -1, dtype=np.int16)
    for i in range(n):
        merged_masks[masks[i] > 0] = i

    return merged_masks



from typing import Generic, List, TypeVar

import depthai as dai

TReference = TypeVar("TReference", bound=dai.Buffer)
TGathered = TypeVar("TGathered")


class GatheredData(dai.Buffer, Generic[TReference, TGathered]):
    """Contains N messages and reference data that the messages were matched with.

    Attributes
    ----------
    reference_data: TReference
        Data that is used to determine how many of TGathered to gather.
    collected: List[TGathered]
        List of collected data.
    """

    def __init__(self, reference_data: TReference, gathered: List[TGathered]) -> None:
        """Initializes the GatheredData object."""
        super().__init__()
        self.reference_data = reference_data
        self.gathered = gathered

    @property
    def reference_data(self) -> TReference:
        """Returns the reference data.

        @return: Reference data.
        @rtype: TReference
        """
        return self._reference_data

    @reference_data.setter
    def reference_data(self, value: TReference):
        """Sets the reference data.

        @param value: Reference data.
        @type value: TReference
        """
        self.setSequenceNum(value.getSequenceNum())
        self.setTimestamp(value.getTimestamp())
        self.setTimestampDevice(value.getTimestampDevice())
        self._reference_data = value

    @property
    def gathered(self) -> List[TGathered]:
        """Returns the collected data.

        @return: List of collected data.
        @rtype: List[TGathered]
        """
        return self._gathered

    @gathered.setter
    def gathered(self, value: List[TGathered]):
        """Sets the gathered data.

        @param value: List of gathered data.
        @type value: List[TGathered]
        @raise TypeError: If value is not a list.
        """
        if not isinstance(value, list):
            raise TypeError("gathered_data must be a list.")
        self._gathered = value



from typing import Any, Dict, List, Optional, Tuple

import depthai as dai
import numpy as np

from depthai_nodes.message.creators import create_keypoints_message
from depthai_nodes.node.parsers.base_parser import BaseParser


class KeypointParser(BaseParser):
    """Parser class for 2D or 3D keypoints models. It expects one ouput layer containing
    keypoints. The number of keypoints must be specified. Moreover, the keypoints are
    normalized by a scale factor if provided.

    Attributes
    ----------
    output_layer_name: str
        Name of the output layer relevant to the parser.
    scale_factor : float
        Scale factor to divide the keypoints by.
    n_keypoints : int
        Number of keypoints the model detects.
    score_threshold : float
        Confidence score threshold for detected keypoints.
    label_names : List[str]
        Label names for the keypoints.
    edges : List[Tuple[int, int]]
        Keypoint connection pairs for visualizing the skeleton. Example: [(0,1), (1,2), (2,3), (3,0)] shows that keypoint 0 is connected to keypoint 1, keypoint 1 is connected to keypoint 2, etc.

    Output Message/s
    ----------------
    **Type**: Keypoints

    **Description**: Keypoints message containing 2D or 3D keypoints.

    Error Handling
    --------------
    **ValueError**: If the number of keypoints is not specified.

    **ValueError**: If the number of coordinates per keypoint is not 2 or 3.

    **ValueError**: If the number of output layers is not 1.
    """

    def __init__(
        self,
        output_layer_name: str = "",
        scale_factor: float = 1.0,
        n_keypoints: int = None,
        score_threshold: float = None,
        label_names: Optional[List[str]] = None,
        edges: Optional[List[List[int]]] = None,
    ) -> None:
        """Initializes the parser node.

        @param output_layer_name: Name of the output layer relevant to the parser.
        @type output_layer_name: str
        @param scale_factor: Scale factor to divide the keypoints by.
        @type scale_factor: float
        @param n_keypoints: Number of keypoints.
        @type n_keypoints: int
        @param label_names: Label names for the keypoints.
        @type label_names: Optional[List[str]]
        @param edges: Keypoint connection pairs for visualizing the skeleton. Example:
            [(0,1), (1,2), (2,3), (3,0)] shows that keypoint 0 is connected to keypoint
            1, keypoint 1 is connected to keypoint 2, etc.
        @type edges: Optional[List[Tuple[int, int]]]
        """
        super().__init__()
        self.output_layer_name = output_layer_name
        self.scale_factor = scale_factor
        self.n_keypoints = n_keypoints
        self.score_threshold = score_threshold
        self.label_names = label_names
        self.edges = edges
        self._logger.debug(
            f"KeypointParser initialized with output_layer_name='{output_layer_name}', scale_factor={scale_factor}, n_keypoints={n_keypoints}, score_threshold={score_threshold}, label_names={label_names}, edges={edges}"
        )

    def setOutputLayerName(self, output_layer_name: str) -> None:
        """Sets the name of the output layer.

        @param output_layer_name: The name of the output layer.
        @type output_layer_name: str
        """
        if not isinstance(output_layer_name, str):
            raise ValueError("Output layer name must be a string.")
        self.output_layer_name = output_layer_name
        self._logger.debug(f"Output layer name set to '{self.output_layer_name}'")

    def setScaleFactor(self, scale_factor: float) -> None:
        """Sets the scale factor to divide the keypoints by.

        @param scale_factor: Scale factor to divide the keypoints by.
        @type scale_factor: float
        """
        if not isinstance(scale_factor, float):
            raise ValueError("Scale factor must be a float.")

        if scale_factor <= 0:
            raise ValueError("Scale factor must be greater than 0.")

        self.scale_factor = scale_factor
        self._logger.debug(f"Scale factor set to {self.scale_factor}")

    def setNumKeypoints(self, n_keypoints: int) -> None:
        """Sets the number of keypoints.

        @param n_keypoints: Number of keypoints.
        @type n_keypoints: int
        """
        if not isinstance(n_keypoints, int):
            raise ValueError("Number of keypoints must be an integer.")

        if n_keypoints <= 0:
            raise ValueError("Number of keypoints must be greater than 0.")

        self.n_keypoints = n_keypoints
        self._logger.debug(f"Number of keypoints set to {self.n_keypoints}")

    def setScoreThreshold(self, threshold: float) -> None:
        """Sets the confidence score threshold for the detected body keypoints.

        @param threshold: Confidence score threshold for detected keypoints.
        @type threshold: float
        """
        if not isinstance(threshold, float):
            raise ValueError("Confidence threshold must be a float.")

        if threshold < 0 or threshold > 1:
            raise ValueError("Confidence threshold must be between 0 and 1.")

        self.score_threshold = threshold
        self._logger.debug(f"Score threshold set to {self.score_threshold}")

    def setLabelNames(self, label_names: List[str]) -> None:
        """Sets the label names for the keypoints.

        @param label_names: List of label names for the keypoints.
        @type label_names: List[str]
        """
        if not isinstance(label_names, list):
            raise ValueError("Label names must be a list.")
        if not all(isinstance(label, str) for label in label_names):
            raise ValueError("Label names must be a list of strings.")
        self.label_names = label_names
        self._logger.debug(f"Label names set to {self.label_names}")

    def setEdges(self, edges: List[Tuple[int, int]]) -> None:
        """Sets the edges for the keypoints.

        @param edges: List of edges for the keypoints. Example: [(0,1), (1,2), (2,3),
            (3,0)] shows that keypoint 0 is connected to keypoint 1, keypoint 1 is
            connected to keypoint 2, etc.
        @type edges: List[Tuple[int, int]]
        """
        if not isinstance(edges, list):
            raise ValueError("Edges must be a list.")
        if not all(
            isinstance(edge, tuple)
            and len(edge) == 2
            and all(isinstance(i, int) for i in edge)
            for edge in edges
        ):
            raise ValueError("Edges must be a list of tuples of integers.")
        self.edges = edges
        self._logger.debug(f"Edges set to {self.edges}")

    def build(
        self,
        head_config: Dict[str, Any],
    ) -> "KeypointParser":
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: KeypointParser
        """

        output_layers = head_config["outputs"]
        if len(output_layers) != 1:
            raise ValueError(
                f"Only one output layer supported for Keypoint, got {output_layers} layers."
            )
        self.output_layer_name = output_layers[0]
        self.scale_factor = head_config.get("scale_factor", self.scale_factor)
        self.n_keypoints = head_config.get("n_keypoints", self.n_keypoints)
        self.score_threshold = head_config.get("score_threshold", self.score_threshold)
        self.label_names = head_config.get("keypoint_labels", self.label_names)
        keypoint_edges = head_config.get("skeleton_edges", self.edges)
        if keypoint_edges:
            self.edges = [tuple(edge) for edge in keypoint_edges]

        self._logger.debug(
            f"KeypointParser built with output_layer_name='{self.output_layer_name}', scale_factor={self.scale_factor}, n_keypoints={self.n_keypoints}, score_threshold={self.score_threshold}, label_names={self.label_names}, edges={self.edges}"
        )

        return self

    def run(self):
        self._logger.debug("KeypointParser run started")
        if self.n_keypoints is None:
            raise ValueError("Number of keypoints must be specified!")

        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            layers = output.getAllLayerNames()
            self._logger.debug(f"Processing input with layers: {layers}")
            if len(layers) == 1 and self.output_layer_name == "":
                self.output_layer_name = layers[0]
            elif len(layers) != 1 and self.output_layer_name == "":
                raise ValueError(
                    f"Expected 1 output layer, got {len(layers)} layers. Please provide the output_layer_name."
                )

            keypoints = output.getTensor(
                self.output_layer_name, dequantize=True
            ).astype(np.float32)
            num_coords = int(np.prod(keypoints.shape) / self.n_keypoints)

            if num_coords not in [2, 3]:
                raise ValueError(
                    f"Expected 2 or 3 coordinates per keypoint, got {num_coords}."
                )

            keypoints = keypoints.reshape(self.n_keypoints, num_coords)

            keypoints /= self.scale_factor

            keypoints = np.clip(keypoints, 0, 1)

            msg = create_keypoints_message(
                keypoints, edges=self.edges, label_names=self.label_names
            )
            msg.setTimestamp(output.getTimestamp())
            msg.setTransformation(output.getTransformation())
            msg.setSequenceNum(output.getSequenceNum())

            self._logger.debug(
                f"Created keypoints message with {len(keypoints)} points"
            )

            self.out.send(msg)

            self._logger.debug("Keypoints message sent successfully")



from typing import Union

import depthai as dai

from .input import InputMock
from .output import OutputMock


class NeuralNetworkMock:
    def __init__(self):
        self._input = InputMock()
        self._out = OutputMock()
        self._passthrough = OutputMock()
        self._nn_archive = None

    def build(
        self,
        input: InputMock,
        model: Union[dai.NNModelDescription, dai.NNArchive],
        fps: float,
    ):
        self._nn_archive = model
        self._input = input
        return self



import time
from typing import Union

import depthai as dai
import pytest
from pytest import FixtureRequest

from depthai_nodes import ImgDetectionsExtended
from depthai_nodes.node import ImgDetectionsFilter
from tests.utils import (
    DETECTIONS,
    LOG_INTERVAL,
    OutputMock,
    create_img_detections,
    create_img_detections_extended,
)

LABELS = [1]
CONF_THRES = 0.5
MAX_DET = 1
SORT = True

img_det_types = ["img_detections", "img_detections_extended"]


@pytest.fixture(scope="session")
def duration(request):
    d = request.config.getoption("--duration")
    if d is None:
        return 1e-6
    return d


@pytest.fixture
def filter():
    return ImgDetectionsFilter()


@pytest.fixture
def img_detections():
    return create_img_detections()


@pytest.fixture
def img_detections_extended():
    return create_img_detections_extended()


def test_initialization(filter: ImgDetectionsFilter):
    assert filter._labels_to_keep is None
    assert filter._labels_to_reject is None
    assert filter._confidence_threshold is None
    assert filter._max_detections is None
    assert filter._sort_by_confidence is False


def test_building(filter: ImgDetectionsFilter):
    filter.build(OutputMock())
    assert filter._labels_to_keep is None
    assert filter._labels_to_reject is None
    assert filter._confidence_threshold is None
    assert filter._max_detections is None
    assert filter._sort_by_confidence is False

    # labels to keep
    filter.build(OutputMock(), labels_to_keep=LABELS)
    assert filter._labels_to_keep == LABELS
    assert filter._labels_to_reject is None

    # labels to reject
    filter.build(
        OutputMock(),
        labels_to_reject=LABELS,
    )
    assert filter._labels_to_keep is None
    assert filter._labels_to_reject == LABELS

    # labels_to_keep and labels_to_reject cannot be set at the same time
    with pytest.raises(ValueError):
        filter.build(
            OutputMock(),
            labels_to_keep=LABELS,
            labels_to_reject=LABELS,
        )

    # confidence_threshold
    filter.build(
        OutputMock(),
        confidence_threshold=CONF_THRES,
    )
    assert filter._confidence_threshold == CONF_THRES

    # max_detections
    filter.build(
        OutputMock(),
        max_detections=MAX_DET,
    )
    assert filter._max_detections == MAX_DET

    # sort_by_confidence
    filter.build(
        OutputMock(),
        sort_by_confidence=SORT,
    )
    assert filter._sort_by_confidence == SORT


def test_parameter_setting(filter: ImgDetectionsFilter):
    filter.build(OutputMock())

    # labels
    filter.setLabels(LABELS, keep=True)
    assert filter._labels_to_keep == LABELS
    filter.setLabels(LABELS, keep=False)
    assert filter._labels_to_keep is None
    assert filter._labels_to_reject == LABELS
    with pytest.raises(ValueError):
        filter.setLabels("not a list", keep=True)
    with pytest.raises(ValueError):
        filter.setLabels(LABELS, keep="not a boolean")

    # confidence_threshold
    filter.setConfidenceThreshold(CONF_THRES)
    assert filter._confidence_threshold == CONF_THRES
    with pytest.raises(ValueError):
        filter.setConfidenceThreshold("not a float")

    # max_detections
    filter.setMaxDetections(MAX_DET)
    assert filter._max_detections == MAX_DET
    with pytest.raises(ValueError):
        filter.setMaxDetections("not an integer")

    # sort_by_confidence
    filter.setSortByConfidence(SORT)
    assert filter._sort_by_confidence == SORT
    with pytest.raises(ValueError):
        filter.setSortByConfidence("not a boolean")


@pytest.mark.parametrize(
    "img_detections_type",
    img_det_types,
)
def test_processing(
    filter: ImgDetectionsFilter,
    request: FixtureRequest,
    img_detections_type: str,
    duration: float,
):
    dets: Union[ImgDetectionsExtended, dai.ImgDetections] = request.getfixturevalue(
        img_detections_type
    )

    o_dets = OutputMock()
    filter.build(o_dets)
    q_dets = o_dets.createOutputQueue()
    q_dets_filtered = filter.out.createOutputQueue()

    modified_duration = duration / len(img_det_types)

    start_time = time.time()
    last_log_time = time.time()
    while time.time() - start_time < modified_duration:
        if time.time() - last_log_time > LOG_INTERVAL:
            print(
                f"Test running... {time.time()-start_time:.1f}s elapsed, {modified_duration-time.time()+start_time:.1f}s remaining"
            )
            last_log_time = time.time()
        # default filtering
        q_dets.send(dets)
        filter.process(q_dets.get())
        dets_filtered = q_dets_filtered.get()
        assert isinstance(dets_filtered, type(dets))
        assert len(dets_filtered.detections) == len(DETECTIONS["bboxes"])

        # filter by labels_to_keep
        filter.setLabels(LABELS, keep=True)
        q_dets.send(dets)
        filter.process(q_dets.get())
        dets_filtered = q_dets_filtered.get()
        assert len(dets_filtered.detections) == len(
            [None for label in DETECTIONS["labels"] if label in LABELS]
        )
        for det in dets_filtered.detections:
            assert det.label in LABELS
        filter.setLabels(None, keep=True)  # setting back to default

        # filter by labels_to_reject
        filter.setLabels(LABELS, keep=False)
        q_dets.send(dets)
        filter.process(q_dets.get())
        dets_filtered = q_dets_filtered.get()
        assert len(dets_filtered.detections) == len(
            [None for label in DETECTIONS["labels"] if label not in LABELS]
        )
        for det in dets_filtered.detections:
            assert det.label not in LABELS
        filter.setLabels(None, keep=False)  # setting back to default

        # filter by confidence
        filter.setConfidenceThreshold(CONF_THRES)
        q_dets.send(dets)
        filter.process(q_dets.get())
        dets_filtered = q_dets_filtered.get()
        assert len(dets_filtered.detections) == len(
            [None for score in DETECTIONS["scores"] if score >= CONF_THRES]
        )
        for det in dets_filtered.detections:
            assert det.confidence >= CONF_THRES
        filter.setConfidenceThreshold(None)  # setting back to default

        # filter by max detections
        filter.setMaxDetections(MAX_DET)
        q_dets.send(dets)
        filter.process(q_dets.get())
        dets_filtered = q_dets_filtered.get()
        assert len(dets_filtered.detections) == MAX_DET
        filter.setMaxDetections(None)  # setting it back to default

        # sort by confidence
        filter.setSortByConfidence(SORT)
        q_dets.send(dets)
        filter.process(q_dets.get())
        dets_filtered = q_dets_filtered.get()
        if SORT:
            assert dets_filtered.detections == sorted(
                dets_filtered.detections, key=lambda x: x.confidence, reverse=True
            )
        filter.setSortByConfidence(False)



import os
import sys
from typing import Dict, List, Tuple

import depthai as dai
import requests

API_KEY = os.getenv("HUBAI_API_KEY", None)
HUBAI_TEAM_SLUG = os.getenv("HUBAI_TEAM_SLUG", None)

if not API_KEY:
    raise ValueError(
        "You must specify your HubAI API key (HUBAI_API_KEY) in order to get the model config."
    )

if not HUBAI_TEAM_SLUG:
    raise ValueError(
        "You must specify your HubAI team slug (HUBAI_TEAM_SLUG) in order to get the model config."
    )

HEADERS = {"Authorization": f"Bearer {API_KEY}"}


def get_inputs_from_archive(nn_archive: dai.NNArchive) -> List:
    """Get all inputs from NN archive."""
    try:
        inputs = nn_archive.getConfig().model.inputs
    except AttributeError:
        print(
            "This NN archive does not have an input shape. Please use NN archives that have input shapes."
        )
        exit(1)

    return inputs


def get_input_shape(nn_archive: dai.NNArchive) -> Tuple[int, int]:
    """Get the input shape of the model from the NN archive."""
    inputs = get_inputs_from_archive(nn_archive)

    if len(inputs) > 1:
        raise ValueError(
            "This model has more than one input. Currently, only models with one input are supported."
        )

    try:
        if inputs[0].layout == "NCHW":
            input_shape = inputs[0].shape[2:][::-1]
        elif inputs[0].layout == "NHWC":
            input_shape = inputs[0].shape[1:3][::-1]
        else:
            raise ValueError(
                "This model has an unsupported layout. Currently, only NCHW and NHWC layouts are supported."
            )
    except AttributeError:
        print(
            "This NN archive does not have an input shape. Please use NN archives that have input shapes."
        )
        exit(1)

    return input_shape


def get_num_inputs(nn_archive: dai.NNArchive) -> int:
    """Get the number of inputs of the model from the NN archive."""
    inputs = get_inputs_from_archive(nn_archive)

    return len(inputs)


def get_models() -> List[Dict]:
    """Get all the models from the ZOO that correspond to the HubAI team."""
    url = "https://easyml.cloud.luxonis.com/models/api/v1/models?is_public=true&limit=1000"

    response = requests.get(url, headers=HEADERS)

    if response.status_code != 200:
        raise ValueError(f"Failed to get models. Status code: {response.status_code}")
    response = response.json()

    valid_models = []

    for model in response:
        if model["is_public"] and model["team_slug"] == HUBAI_TEAM_SLUG:
            model_dict = {
                "name": model["name"],
                "slug": model["slug"],
                "model_id": model["id"],
            }
            valid_models.append(model_dict)

    return valid_models


def get_model_versions(models: List[Dict]) -> List[Dict]:
    """Get all the model versions from the ZOO that correspond to given models."""
    url = "https://easyml.cloud.luxonis.com/models/api/v1/modelVersions?is_public=True&limit=1000"

    response = requests.get(url, headers=HEADERS)

    if response.status_code != 200:
        print(f"Failed to get model versions. Status code: {response.status_code}")
        return None

    response = response.json()

    model_versions = []

    for version in response:
        for model in models:
            if version["model_id"] == model["model_id"]:
                model_version = {
                    "slug": model["slug"],
                    "version_slug": version["variant_slug"],
                    "model_id": model["model_id"],
                    "version_id": version["id"],
                    "name": model["name"],
                }
                model_versions.append(model_version)
                break

    return model_versions


def get_model_instances(model_versions: List[Dict]) -> List[Dict]:
    """Get all the model instances from the ZOO that correspond to the model
    versions."""
    url = "https://easyml.cloud.luxonis.com/models/api/v1/modelInstances?is_public=True&limit=1000"
    response = requests.get(url, headers=HEADERS)

    if response.status_code != 200:
        print(f"Failed to get model instances. Status code: {response.status_code}")
        return None

    response = response.json()

    model_instances = []

    already_checked = []

    for instance in response:
        for model in model_versions:
            if instance["model_version_id"] == model["version_id"]:
                if instance["model_version_id"] in already_checked:
                    continue
                model_instance = {
                    "slug": model["slug"],
                    "version_slug": model["version_slug"],
                    "model_id": model["model_id"],
                    "version_id": model["version_id"],
                    "instance_id": instance["id"],
                }
                model_instances.append(model_instance)
                already_checked.append(instance["model_version_id"])
                break

    return model_instances


def get_model_config_parser(model_instance_id: str) -> Dict:
    """Get the model config from the ZOO."""
    url = f"https://easyml.cloud.luxonis.com/models/api/v1/modelInstances/{model_instance_id}/config"
    response = requests.get(url, headers=HEADERS)

    if response.status_code != 200:
        print(f"Failed to get model config. Status code: {response.status_code}")
        return None

    return response.json()


def get_model_slugs_from_zoo() -> List[str]:
    """Get all the model slugs from the ZOO."""
    model_slugs = []
    models = get_models()
    model_versions = get_model_versions(models)

    for model in model_versions:
        model_slugs.append(f"{model['slug']}:{model['version_slug']}")

    return model_slugs


def find_slugs_from_zoo(parser: str) -> List[str]:
    """Find all model slugs that have the required parser."""
    models = get_models()
    model_versions = get_model_versions(models)
    model_instances = get_model_instances(model_versions)
    relevant_slugs = []
    already_checked = []

    for ix, instance in enumerate(model_instances):
        sys.stdout.write(
            f"\r[{ix + 1}/{len(model_instances)}] Checking model configs..."
        )
        sys.stdout.flush()
        model_config = get_model_config_parser(instance["instance_id"])
        if model_config is None:
            continue
        try:
            heads = model_config["model"]["heads"]
        except KeyError:
            print(
                f"Model {instance['slug']}:{instance['version_slug']} does not have heads."
            )
            continue

        try:
            parsers = [head["parser"] for head in heads]
        except KeyError:
            print(
                f"Model {instance['slug']}:{instance['version_slug']} does not have a parser."
            )
            continue

        if parser in parsers:
            if instance["version_id"] in already_checked:
                continue
            relevant_slugs.append(f"{instance['slug']}:{instance['version_slug']}")
            already_checked.append(instance["version_id"])

    print()
    return relevant_slugs



from datetime import timedelta
from typing import List, Optional, Tuple

import depthai as dai
import numpy as np

from depthai_nodes.constants import (
    PRIMARY_COLOR,
    TRANSPARENT_PRIMARY_COLOR,
)

Point = Tuple[float, float]
ColorRGBA = Tuple[float, float, float, float]


class AnnotationHelper:
    """Simplifies `dai.ImgAnnotation` creation.

    After calling the desired drawing methods, call the `build` method to create the `ImgAnnotations` message.
    """

    def __init__(self):
        self.annotation: dai.ImgAnnotation = dai.ImgAnnotation()

    def draw_line(
        self,
        pt1: Point,
        pt2: Point,
        color: ColorRGBA = PRIMARY_COLOR,
        thickness: float = 2,
    ) -> "AnnotationHelper":
        """Draws a line between two points.

        @param pt1: Start of the line
        @type pt1: Point
        @param pt2: End of the line
        @type pt2: Point
        @param color: Line color
        @type color: ColorRGBA
        @param thickness: Line thickness
        @type thickness: float
        @return: self
        @rtype: AnnotationHelper
        """
        line = dai.PointsAnnotation()
        c = self._create_color(color)
        line.fillColor = c
        line.outlineColor = c
        line.thickness = thickness
        line.type = dai.PointsAnnotationType.LINE_STRIP
        line.points = self._create_points_vector([pt1, pt2])
        self.annotation.points.append(line)
        return self

    def draw_polyline(
        self,
        points: List[Point],
        outline_color: ColorRGBA = PRIMARY_COLOR,
        fill_color: Optional[ColorRGBA] = TRANSPARENT_PRIMARY_COLOR,
        thickness: float = 1,
        closed: bool = False,
    ) -> "AnnotationHelper":
        """Draws a polyline.

        @param points: List of points of the polyline
        @type points: list[Point]
        @param outline_color: Outline color
        @type outline_color: ColorRGBA
        @param fill_color: Fill color (None for no fill), defaults to None
        @type fill_color: ColorRGBA | None, optional
        @param thickness: Line thickness, defaults to 1
        @type thickness: float, optional
        @param closed: Creates polygon, instead of polyline if True, defaults to False
        @type closed: bool, optional
        @return: self
        @rtype: AnnotationHelper
        """
        points_type = (
            dai.PointsAnnotationType.LINE_STRIP
            if not closed
            else dai.PointsAnnotationType.LINE_LOOP
        )
        if not closed:
            fill_color = None
        points_annot = self._create_points_annotation(
            points, outline_color, fill_color, points_type
        )
        points_annot.thickness = thickness
        self.annotation.points.append(points_annot)
        return self

    def draw_points(
        self,
        points: List[Point],
        color: ColorRGBA = PRIMARY_COLOR,
        thickness: float = 2,
    ) -> "AnnotationHelper":
        """Draws points.

        @param points: List of points to draw
        @type points: list[Point]
        @param color: Color of the points
        @type color: ColorRGBA
        @param thickness: Size of the points, defaults to 2
        @type thickness: float, optional
        @return: self
        @rtype: AnnotationHelper
        """
        # TODO: Visualizer currently does not show dai.PointsAnnotationType.POINTS
        points_annot = self._create_points_annotation(
            points, color, None, dai.PointsAnnotationType.POINTS
        )
        points_annot.thickness = thickness
        self.annotation.points.append(points_annot)
        return self

    def draw_circle(
        self,
        center: Point,
        radius: float,
        outline_color: ColorRGBA = PRIMARY_COLOR,
        fill_color: Optional[ColorRGBA] = None,
        thickness: float = 1,
    ) -> "AnnotationHelper":
        """Draws a circle.

        @param center: Center of the circle
        @type center: Point
        @param radius: Radius of the circle
        @type radius: float
        @param outline_color: Outline color
        @type outline_color: ColorRGBA
        @param fill_color: Fill color (None for no fill), defaults to None
        @type fill_color: ColorRGBA | None, optional
        @param thickness: Outline thickness, defaults to 1
        @type thickness: float, optional
        @return: self
        @rtype: AnnotationHelper
        """
        circle = dai.CircleAnnotation()
        circle.outlineColor = self._create_color(outline_color)
        if fill_color is not None:
            circle.fillColor = self._create_color(fill_color)
        circle.thickness = thickness
        circle.diameter = radius * 2
        circle.position = dai.Point2f(center[0], center[1])
        self.annotation.circles.append(circle)
        return self

    def draw_rectangle(
        self,
        top_left: Point,
        bottom_right: Point,
        outline_color: ColorRGBA = PRIMARY_COLOR,
        fill_color: Optional[ColorRGBA] = TRANSPARENT_PRIMARY_COLOR,
        thickness: float = 1,
    ) -> "AnnotationHelper":
        """Draws a rectangle.

        @param top_left: Top left corner of the rectangle
        @type top_left: Point
        @param bottom_right: Bottom right corner of the rectangle
        @type bottom_right: Point
        @param outline_color: Outline color
        @type outline_color: ColorRGBA
        @param fill_color: Fill color (None for no fill), defaults to None
        @type fill_color: ColorRGBA | None, optional
        @param thickness: Outline thickness, defaults to 1
        @type thickness: float, optional
        @return: self
        @rtype: AnnotationHelper
        """
        points = [
            top_left,
            (bottom_right[0], top_left[1]),
            bottom_right,
            (top_left[0], bottom_right[1]),
        ]
        self.draw_polyline(points, outline_color, fill_color, thickness, closed=True)
        return self

    def draw_text(
        self,
        text: str,
        position: Point,
        color: ColorRGBA = PRIMARY_COLOR,
        background_color: Optional[ColorRGBA] = None,
        size: float = 32,
    ) -> "AnnotationHelper":
        """Draws text.

        @param text: Text string
        @type text: str
        @param position: Text position
        @type position: Point
        @param color: Text color
        @type color: ColorRGBA
        @param background_color: Background color (None for no background), defaults to
            None
        @type background_color: ColorRGBA | None, optional
        @param size: Text size, defaults to 32
        @type size: float, optional
        @return: self
        @rtype: AnnotationHelper
        """
        text_annot = dai.TextAnnotation()
        text_annot.position = dai.Point2f(position[0], position[1])
        text_annot.text = text
        text_annot.textColor = self._create_color(color)
        text_annot.fontSize = size
        if background_color is not None:
            text_annot.backgroundColor = self._create_color(background_color)
        self.annotation.texts.append(text_annot)
        return self

    def draw_rotated_rect(
        self,
        center: Point,
        size: Tuple[float, float],
        angle: float,
        outline_color: ColorRGBA = PRIMARY_COLOR,
        fill_color: Optional[ColorRGBA] = TRANSPARENT_PRIMARY_COLOR,
        thickness: float = 1,
    ) -> "AnnotationHelper":
        """Draws a rotated rectangle.

        @param center: Center of the rectangle
        @type center: Point
        @param size: Size of the rectangle (width, height)
        @type size: tuple[float, float]
        @param angle: Angle of rotation in degrees
        @type angle: float
        @param outline_color: Outline color
        @type outline_color: ColorRGBA
        @param fill_color: Fill color (None for no fill), defaults to None
        @type fill_color: ColorRGBA | None, optional
        @param thickness: Outline thickness, defaults to 1
        @type thickness: float, optional
        @return: self
        @rtype: AnnotationHelper
        """
        points = self._get_rotated_rect_points(center, size, angle)
        self.draw_polyline(points, outline_color, fill_color, thickness, True)
        return self

    def build(self, timestamp: timedelta, sequence_num: int) -> dai.ImgAnnotations:
        """Creates an ImgAnnotations message.

        @param timestamp: Message timestamp
        @type timestamp: timedelta
        @param sequence_num: Message sequence number
        @type sequence_num: int
        @return: Created ImgAnnotations message
        @rtype: dai.ImgAnnotations
        """
        annotations_msg = dai.ImgAnnotations()
        annotations_msg.annotations = dai.VectorImgAnnotation([self.annotation])
        annotations_msg.setTimestamp(timestamp)
        annotations_msg.setSequenceNum(sequence_num)
        return annotations_msg

    def _create_points_annotation(
        self,
        points: List[Point],
        color: ColorRGBA,
        fill_color: Optional[ColorRGBA],
        type: dai.PointsAnnotationType,
    ) -> dai.PointsAnnotation:
        points_annot = dai.PointsAnnotation()
        points_annot.outlineColor = self._create_color(color)
        if fill_color is not None:
            points_annot.fillColor = self._create_color(fill_color)
        points_annot.type = type
        points_annot.points = self._create_points_vector(points)
        return points_annot

    def _create_color(self, color: ColorRGBA) -> dai.Color:
        if isinstance(color, dai.Color):
            return color
        c = dai.Color()
        c.a = color[3]
        c.r = color[0]
        c.g = color[1]
        c.b = color[2]
        return c

    def _get_rotated_rect_points(
        self, center: Point, size: Tuple[float, float], angle: float
    ) -> List[Point]:
        cx, cy = center
        width, height = size
        angle_rad = np.radians(angle)

        # Half-dimensions
        dx = width / 2
        dy = height / 2

        # Define the corners relative to the center
        corners = np.array([[-dx, -dy], [dx, -dy], [dx, dy], [-dx, dy]])

        # Rotation matrix
        rotation_matrix = np.array(
            [
                [np.cos(angle_rad), -np.sin(angle_rad)],
                [np.sin(angle_rad), np.cos(angle_rad)],
            ]
        )

        # Rotate and translate the corners
        rotated_corners = corners @ rotation_matrix.T
        translated_corners = rotated_corners + np.array([cx, cy])

        # Convert to list of tuples
        return translated_corners.tolist()

    def _create_points_vector(self, points: List[Point]) -> dai.VectorPoint2f:
        return dai.VectorPoint2f([dai.Point2f(pt[0], pt[1]) for pt in points])



from typing import List, Optional, Tuple

import numpy as np

from depthai_nodes import (
    ImgDetectionExtended,
    ImgDetectionsExtended,
)

from .keypoints import create_keypoints_message


def create_detection_message(
    bboxes: np.ndarray,
    scores: np.ndarray,
    angles: np.ndarray = None,
    labels: np.ndarray = None,
    label_names: Optional[List[str]] = None,
    keypoints: np.ndarray = None,
    keypoints_scores: np.ndarray = None,
    keypoint_label_names: Optional[List[str]] = None,
    keypoint_edges: Optional[List[Tuple[int, int]]] = None,
    masks: np.ndarray = None,
) -> ImgDetectionsExtended:
    """Create a DepthAI message for object detection. The message contains the bounding
    boxes in X_center, Y_center, Width, Height format with optional angles, labels and
    detected object keypoints and masks.

    @param bbox: Bounding boxes of detected objects in (x_center, y_center, width,
        height) format.
    @type bbox: np.ndarray
    @param scores: Confidence scores of the detected objects of shape (N,).
    @type scores: np.ndarray
    @param angles: Angles of detected objects expressed in degrees. Defaults to None.
    @type angles: Optional[np.ndarray]
    @param labels: Labels of detected objects of shape (N,). Defaults to None.
    @type labels: Optional[np.ndarray]
    @param label_names: Names of the labels (classes)
    @type label_names: Optional[List[str]]
    @param keypoints: Keypoints of detected objects of shape (N, n_keypoints, dim) where
        dim is 2 or 3. Defaults to None.
    @type keypoints: Optional[np.array]
    @param keypoints_scores: Confidence scores of detected keypoints of shape (N,
        n_keypoints). Defaults to None.
    @type keypoints_scores: Optional[np.ndarray]
    @param keypoint_label_names: Labels of keypoints. Defaults to None.
    @type keypoint_label_names: Optional[List[str]]
    @param keypoint_edges: Connection pairs of keypoints. Defaults to None. Example:
        [(0,1), (1,2), (2,3), (3,0)] shows that keypoint 0 is connected to keypoint 1,
        keypoint 1 is connected to keypoint 2, etc.
    @type keypoint_edges: Optional[List[Tuple[int, int]]]
    @param masks: Masks of detected objects of shape (H, W). Defaults to None.
    @type masks: Optional[np.ndarray]
    @return: Message containing the bounding boxes, labels, confidence scores, and
        keypoints of detected objects.
    @rtype: ImgDetectionsExtended
    @raise ValueError: If the bboxes are not a numpy array.
    @raise ValueError: If the bboxes are not of shape (N,4).
    @raise ValueError: If the scores are not a numpy array.
    @raise ValueError: If the scores are not of shape (N,).
    @raise ValueError: If the scores do not have the same length as bboxes.
    @raise ValueError: If the angles do not have the same length as bboxes.
    @raise ValueError: If the angles are not between -360 and 360.
    @raise ValueError: If the labels are not a list of integers.
    @raise ValueError: If the labels do not have the same length as bboxes.
    @raise ValueError: If the keypoints are not a numpy array of shape (N, M, 2 or 3).
    @raise ValueError: If the masks are not a 3D numpy array of shape (img_height,
        img_width, N) or (N, img_height, img_width).
    @raise ValueError: If the keypoints scores are not a numpy array.
    @raise ValueError: If the keypoints scores are not of shape [n_detections,
        n_keypoints, 1].
    @raise ValueError: If the keypoints scores do not have the same length as keypoints.
    @raise ValueError: If the keypoints scores are not between 0 and 1.
    """

    if not isinstance(bboxes, np.ndarray):
        raise ValueError(f"Bounding boxes should be a numpy array, got {type(bboxes)}.")

    if len(bboxes) == 0:
        img_detections = ImgDetectionsExtended()
        if masks is not None:
            img_detections.masks = masks
        return img_detections

    if len(bboxes.shape) != 2:
        raise ValueError(
            f"Bounding boxes should be a 2D array like [... , [x_center, y_center, width, height], ... ], got {bboxes.shape}."
        )
    if bboxes.shape[1] != 4:
        raise ValueError(
            f"Bounding boxes should be of shape (n, 4), got {bboxes.shape}."
        )

    if not isinstance(scores, np.ndarray):
        raise ValueError(f"Scores should be a numpy array, got {type(scores)}.")

    if len(scores.shape) != 1:
        raise ValueError(f"Scores should be of shape (N,), got {scores.shape}.")

    n_bboxes = len(bboxes)
    if len(scores) != n_bboxes:
        raise ValueError(
            f"Scores should have same length as bboxes, got {len(scores)} scores and {n_bboxes} bounding boxes."
        )

    if labels is not None:
        if not isinstance(labels, np.ndarray):
            raise ValueError(f"Labels should be a numpy array, got {type(labels)}.")
        if labels.shape[0] != n_bboxes:
            raise ValueError(
                f"Labels should have same length as bboxes, got {len(labels)} labels and {n_bboxes} bounding boxes."
            )

    if label_names is not None:
        if not isinstance(label_names, list):
            raise ValueError(f"Label names should be a list, got {type(label_names)}.")
        if not all(isinstance(label_name, str) for label_name in label_names):
            raise ValueError(f"Label names should be strings, got {label_names}.")

    if angles is not None:
        if not isinstance(angles, np.ndarray):
            raise ValueError(f"Angles should be a numpy array, got {type(angles)}.")
        if len(angles) != n_bboxes:
            raise ValueError(
                f"Angles should have same length as bboxes, got {len(angles)} angles and {n_bboxes} bounding boxes."
            )
        for angle in angles:
            if not -360 <= angle <= 360:
                raise ValueError(f"Angles should be between -360 and 360, got {angle}.")

    if keypoints is not None:
        if not isinstance(keypoints, np.ndarray):
            raise ValueError(
                f"Keypoints should be a numpy array, got {type(keypoints)}."
            )
        if len(keypoints.shape) != 3:
            raise ValueError(
                f"Keypoints should be of shape (N, M, 2 or 3) meaning [..., [x, y] or [x, y, z], ...], got {keypoints.shape}."
            )
        n_detections, n_keypoints, dim = keypoints.shape
        keypoints = np.array(keypoints, dtype=float)
        if n_detections != n_bboxes:
            raise ValueError(
                f"Keypoints should have same length as bboxes, got {n_detections} keypoints and {n_bboxes} bounding boxes."
            )
        if dim not in [2, 3]:
            raise ValueError(
                f"Keypoints should be of dimension 2 or 3 e.g. [x, y] or [x, y, z], got {dim} dimensions."
            )

    if keypoints_scores is not None:
        if keypoints is None:
            raise ValueError(
                "Keypoints scores should be provided only if keypoints are provided."
            )
        if not isinstance(keypoints_scores, np.ndarray):
            raise ValueError(
                f"Keypoints scores should be a numpy array, got {type(keypoints_scores)}."
            )

        n_detections, n_keypoints, _ = keypoints.shape
        if keypoints_scores.shape[0] != n_detections:
            raise ValueError(
                f"Keypoints scores should have same length as keypoints, got {len(keypoints_scores)} keypoints scores and {n_detections} keypoints."
            )

        if keypoints_scores.shape[1] != n_keypoints:
            raise ValueError(
                f"Number of keypoints scores per detection should be the same as number of keypoints per detection, got {keypoints_scores.shape[1]} keypoints scores and {n_keypoints} keypoints."
            )

        if not all(0 <= score <= 1 for score in keypoints_scores.flatten()):
            raise ValueError("Keypoints scores should be between 0 and 1.")

    if keypoint_label_names is not None:
        if not isinstance(keypoint_label_names, list):
            raise ValueError(
                f"Keypoint label names should be a list, got {type(keypoint_label_names)}."
            )
        if not all(isinstance(label, str) for label in keypoint_label_names):
            raise ValueError(
                f"Keypoint label names should be a list of strings, got {keypoint_label_names}."
            )

    if keypoint_edges is not None:
        if not isinstance(keypoint_edges, list):
            raise ValueError(
                f"Keypoint edges should be a list, got {type(keypoint_edges)}."
            )
        if not all(
            isinstance(edge, tuple)
            and len(edge) == 2
            and all(isinstance(i, int) for i in edge)
            for edge in keypoint_edges
        ):
            raise ValueError(
                f"Keypoint edges should be a list of tuples of integers, got {keypoint_edges}."
            )

    if masks is not None:
        if not isinstance(masks, np.ndarray):
            raise ValueError(f"Masks should be a numpy array, got {type(masks)}.")

        if len(masks.shape) != 2:
            raise ValueError(f"Masks should be of shape (H, W), got {masks.shape}.")

        if masks.dtype != np.int16:
            masks = masks.astype(np.int16)

    detections = []
    for detection_idx in range(n_bboxes):
        detection = ImgDetectionExtended()

        x_center, y_center, width, height = bboxes[detection_idx]
        angle = 0
        if angles is not None:
            angle = float(angles[detection_idx])
        detection.rotated_rect = (
            float(x_center),
            float(y_center),
            float(width),
            float(height),
            angle,
        )
        detection.confidence = float(scores[detection_idx])

        if labels is not None:
            detection.label = int(labels[detection_idx])
            if label_names is not None:
                detection.label_name = label_names[detection_idx]
        if keypoints is not None:
            keypoints_msg = create_keypoints_message(
                keypoints=keypoints[detection_idx],
                scores=(
                    None
                    if keypoints_scores is None
                    else keypoints_scores[detection_idx]
                ),
                label_names=keypoint_label_names,
                edges=keypoint_edges,
            )
            detection.keypoints = keypoints_msg
        detections.append(detection)

    detections_msg = ImgDetectionsExtended()
    detections_msg.detections = detections

    if masks is not None:
        detections_msg.masks = masks

    return detections_msg



from typing import Any, Dict, List, Tuple

import depthai as dai
import numpy as np

from depthai_nodes.message.creators import create_cluster_message
from depthai_nodes.node.parsers.base_parser import BaseParser
from depthai_nodes.node.parsers.utils.ufld import decode_ufld


class LaneDetectionParser(BaseParser):
    """
    Parser class for Ultra-Fast-Lane-Detection model. It expects one ouput layer containing the lane detection results.
    It supports two versions of the model: CuLane and TuSimple. Results are representented with clusters of points.

    Attributes
    ----------
    output_layer_name: str
        Name of the output layer relevant to the parser.
    row_anchors : List[int]
        List of row anchors.
    griding_num : int
        Griding number.
    cls_num_per_lane : int
        Number of points per lane.
    input_size : Tuple[int, int]
        Input size (width,height).

    Output Message/s
    ----------------
    **Type**: Clusters
    **Description**: Detected lanes represented as clusters of points.

    Error Handling
    --------------
    **ValueError**: If the row anchors are not specified.
    **ValueError**: If the griding number is not specified.
    **ValueError**: If the number of points per lane is not specified.
    """

    def __init__(
        self,
        output_layer_name: str = "",
        row_anchors: List[int] = None,
        griding_num: int = None,
        cls_num_per_lane: int = None,
        input_size: Tuple[int, int] = None,
    ) -> None:
        """Initializes the lane detection parser node.

        @param output_layer_name: Name of the output layer relevant to the parser.
        @type output_layer_name: str
        @param row_anchors: List of row anchors.
        @type row_anchors: List[int]
        @param griding_num: Griding number.
        @type griding_num: int
        @param cls_num_per_lane: Number of points per lane.
        @type cls_num_per_lane: int
        @param input_size: Input size (width,height).
        @type input_size: Tuple[int, int]
        """
        super().__init__()
        self.output_layer_name = output_layer_name

        self.row_anchors = row_anchors
        self.griding_num = griding_num
        self.cls_num_per_lane = cls_num_per_lane
        self.input_size = input_size
        self._logger.debug(
            f"LaneDetectionParser initialized with output_layer_name='{output_layer_name}', row_anchors={row_anchors}, griding_num={griding_num}, cls_num_per_lane={cls_num_per_lane}, input_size={input_size}"
        )

    def setOutputLayerName(self, output_layer_name: str) -> None:
        """Set the output layer name for the lane detection model.

        @param output_layer_name: Name of the output layer.
        @type output_layer_name: str
        """
        if not isinstance(output_layer_name, str):
            raise ValueError("Output layer name must be a string.")
        self.output_layer_name = output_layer_name
        self._logger.debug(f"Output layer name set to '{self.output_layer_name}'")

    def setRowAnchors(self, row_anchors: List[int]) -> None:
        """Set the row anchors for the lane detection model.

        @param row_anchors: List of row anchors.
        @type row_anchors: List[int]
        """
        if not isinstance(row_anchors, list):
            raise ValueError("Row anchors must be a list.")
        if not all(isinstance(anchor, int) for anchor in row_anchors):
            raise ValueError("Row anchors must be a list of integers.")
        self.row_anchors = row_anchors
        self._logger.debug(f"Row anchors set to {self.row_anchors}")

    def setGridingNum(self, griding_num: int) -> None:
        """Set the griding number for the lane detection model.

        @param griding_num: Griding number.
        @type griding_num: int
        """
        if not isinstance(griding_num, int):
            raise ValueError("Griding number must be an integer.")
        self.griding_num = griding_num
        self._logger.debug(f"Griding number set to {self.griding_num}")

    def setClsNumPerLane(self, cls_num_per_lane: int) -> None:
        """Set the number of points per lane for the lane detection model.

        @param cls_num_per_lane: Number of classes per lane.
        @type cls_num_per_lane: int
        """
        if not isinstance(cls_num_per_lane, int):
            raise ValueError("Number of points per lane must be an integer.")
        self.cls_num_per_lane = cls_num_per_lane
        self._logger.debug(f"Number of points per lane set to {self.cls_num_per_lane}")

    def setInputSize(self, input_size: Tuple[int, int]) -> None:
        """Set the input size for the lane detection model.

        @param input_size: Input size (width,height).
        @type input_size: Tuple[int, int]
        """
        if not isinstance(input_size, tuple):
            raise ValueError("Input size must be a tuple.")
        if len(input_size) != 2:
            raise ValueError("Input size must be a tuple of two integers.")
        if not all(isinstance(size, int) for size in input_size):
            raise ValueError("Input size must be a tuple of integers.")
        self.input_size = input_size
        self._logger.debug(f"Input size set to {self.input_size}")

    def build(
        self,
        head_config: Dict[str, Any],
    ) -> "LaneDetectionParser":
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: LaneDetectionParser
        """

        output_layers = head_config.get("outputs", [])
        if len(output_layers) != 1:
            raise ValueError(
                f"Only one output layer supported for LaneDetectionParser, got {len(output_layers)} layers."
            )
        self.output_layer_name = output_layers[0]
        self.row_anchors = head_config.get("row_anchors", self.row_anchors)
        self.griding_num = head_config.get("griding_num", self.griding_num)
        self.cls_num_per_lane = head_config.get(
            "cls_num_per_lane", self.cls_num_per_lane
        )

        inputs = head_config["model_inputs"]
        if len(inputs) != 1:
            raise ValueError(
                f"Only one input supported for LaneDetectionParser, got {len(inputs)} inputs."
            )
        self.input_shape = inputs[0].get("shape")
        self.layout = inputs[0].get("layout")
        if self.layout == "NHWC":
            self.input_size = (self.input_shape[2], self.input_shape[1])
        elif self.layout == "NCHW":
            self.input_size = (self.input_shape[3], self.input_shape[2])
        else:
            raise ValueError(
                f"Input layout {self.layout} not supported for input_size extraction."
            )

        self._logger.debug(
            f"LaneDetectionParser built with output_layer_name='{self.output_layer_name}', row_anchors={self.row_anchors}, griding_num={self.griding_num}, cls_num_per_lane={self.cls_num_per_lane}, input_size={self.input_size}"
        )

        return self

    def run(self):
        self._logger.debug("LaneDetectionParser run started")
        if self.row_anchors is None:
            raise ValueError("Row anchors must be specified!")
        if self.griding_num is None:
            raise ValueError("Griding number must be specified!")
        if self.cls_num_per_lane is None:
            raise ValueError("Number of points per lane must be specified!")

        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            layers = output.getAllLayerNames()
            self._logger.debug(f"Processing input with layers: {layers}")
            if len(layers) == 1 and self.output_layer_name == "":
                self.output_layer_name = layers[0]
            elif len(layers) != 1 and self.output_layer_name == "":
                raise ValueError(
                    f"Expected 1 output layer, got {len(layers)} layers. Please provide the output_layer_name."
                )

            tensor = output.getTensor(self.output_layer_name, dequantize=True).astype(
                np.float32
            )
            y = tensor[0]

            points = decode_ufld(
                anchors=self.row_anchors,
                griding_num=self.griding_num,
                cls_num_per_lane=self.cls_num_per_lane,
                input_width=self.input_size[0],
                input_height=self.input_size[1],
                y=y,
            )

            msg = create_cluster_message(points)
            msg.setTimestamp(output.getTimestamp())
            msg.setTransformation(output.getTransformation())
            msg.setSequenceNum(output.getSequenceNum())

            self._logger.debug(
                f"Created lane detection message with {len(points)} points"
            )

            self.out.send(msg)

            self._logger.debug("Lane detection message sent successfully")



from typing import List, Optional

import depthai as dai
import numpy as np

from depthai_nodes.message.creators import (
    create_detection_message,
)
from depthai_nodes.node.parsers.utils.bbox_format_converters import xyxy_to_xywh
from depthai_nodes.node.parsers.utils.nms import nms_cv2

from .base_parser import BaseParser


class DetectionParser(BaseParser):
    """Parser class for parsing the output of a "general" detection model. The parser expects the
    output of the model to have two tensors: one for bounding boxes and one for scores.
    Tensor for bboxes should be of shape (N, 4) and scores should be of shape (N,).
    Bboxes are expected to be in the format [xmin, ymin, xmax, ymax]. If this is not the case you can check other parsers
    or create a new one. As the result, the node sends out the detected objects in the form of a message
    containing bounding boxes and confidence scores.

    Attributes
    ----------
    output_layer_name: str
        Name of the output layer relevant to the parser.
    conf_threshold : float
        Confidence score threshold of detected bounding boxes.
    iou_threshold : float
        Non-maximum suppression threshold.
    max_det : int
        Maximum number of detections to keep.
    label_names : List[str]
        List of label names for detected objects.

    Output Message/s
        -------
        **Type**: ImgDetectionsExtended

        **Description**: ImgDetectionsExtended message containing bounding boxes and confidence scores of detected objects.
    ----------------
    """

    def __init__(
        self,
        conf_threshold: float = 0.5,
        iou_threshold: float = 0.5,
        max_det: int = 100,
        label_names: Optional[List[str]] = None,
    ) -> None:
        """Initializes the parser node.

        @param conf_threshold: Confidence score threshold of detected bounding boxes.
        @type conf_threshold: float
        @param iou_threshold: Non-maximum suppression threshold.
        @type iou_threshold: float
        @param max_det: Maximum number of detections to keep.
        @type max_det: int
        """
        super().__init__()
        self.conf_threshold = conf_threshold
        self.iou_threshold = iou_threshold
        self.max_det = max_det
        self.label_names = label_names
        self._logger.debug(
            f"DetectionParser initialized with conf_threshold={conf_threshold}, iou_threshold={iou_threshold}, max_det={max_det}"
        )

    def setConfidenceThreshold(self, threshold: float) -> None:
        """Sets the confidence score threshold for detected objects.

        @param threshold: Confidence score threshold for detected objects.
        @type threshold: float
        """
        if not isinstance(threshold, float):
            raise ValueError("Confidence threshold must be a float.")
        self.conf_threshold = threshold
        self._logger.debug(f"Confidence threshold updated to {threshold}")

    def setIOUThreshold(self, threshold: float) -> None:
        """Sets the non-maximum suppression threshold.

        @param threshold: Non-maximum suppression threshold.
        @type threshold: float
        """
        if not isinstance(threshold, float):
            raise ValueError("IOU threshold must be a float.")
        self.iou_threshold = threshold
        self._logger.debug(f"IoU threshold updated to {threshold}")

    def setMaxDetections(self, max_det: int) -> None:
        """Sets the maximum number of detections to keep.

        @param max_det: Maximum number of detections to keep.
        @type max_det: int
        """
        if not isinstance(max_det, int):
            raise ValueError("Max detections must be an integer.")
        self.max_det = max_det
        self._logger.debug(f"Maximum detections updated to {max_det}")

    def setLabelNames(self, label_names: List[str]) -> None:
        """Sets the label names for detected objects.

        @param label_names: List of label names for detected objects.
        @type label_names: List[str]
        """
        if not isinstance(label_names, list):
            raise ValueError("Label names must be a list.")
        if not all(isinstance(label, str) for label in label_names):
            raise ValueError("Each label name must be a string.")
        self.label_names = label_names
        self._logger.debug(f"Label names updated to: {label_names}")

    def build(self, head_config) -> "DetectionParser":
        """Configures the parser.

        @param head_config: The head configuration for the parser.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: DetectionParser
        """

        self.conf_threshold = head_config.get("conf_threshold", self.conf_threshold)
        self.iou_threshold = head_config.get("iou_threshold", self.iou_threshold)
        self.max_det = head_config.get("max_det", self.max_det)
        self.label_names = head_config.get("classes", self.label_names)

        self._logger.debug(
            f"DetectionParser built with conf_threshold={self.conf_threshold}, iou_threshold={self.iou_threshold}, max_det={self.max_det}"
        )
        return self

    def run(self):
        self._logger.debug("DetectionParser run started")
        while self.isRunning():
            try:
                output: dai.NNData = self.input.get()
            except dai.MessageQueue.QueueException:
                break  # Pipeline was stopped

            layers = output.getAllLayerNames()
            if len(layers) != 2:
                raise ValueError(
                    f"Expected 2 output layers, got {len(layers)} layers. Please use different parser or create a new one."
                )

            bboxes = None
            scores = None

            for layer in layers:
                tensor: np.ndarray = output.getTensor(layer, dequantize=True)
                if tensor.shape[-1] == 4 and len(tensor.shape) != 1:
                    bboxes = tensor
                else:
                    scores = tensor

            bboxes = bboxes.reshape(-1, 4)
            scores = scores.reshape(-1)

            if bboxes is None or scores is None:
                raise ValueError(
                    "Bounding boxes or scores are missing in the output. Please check the NN model."
                )

            indices = nms_cv2(
                bboxes, scores, self.conf_threshold, self.iou_threshold, self.max_det
            )

            if len(indices) > 0:
                bboxes = bboxes[indices]
                scores = scores[indices]

                bboxes = xyxy_to_xywh(bboxes)

            else:
                bboxes = np.array([])
                scores = np.array([])

            message = create_detection_message(bboxes=bboxes, scores=scores)
            message.transformation = output.getTransformation()
            message.setTimestamp(output.getTimestamp())
            message.setSequenceNum(output.getSequenceNum())

            self._logger.debug(f"Created detections message with {len(bboxes)} objects")
            self.out.send(message)
            self._logger.debug("Detections message sent successfully")



import argparse
import json
import os
import sys

import pytest
from utils import find_slugs_from_zoo, get_model_slugs_from_zoo

from depthai_nodes.logging import get_logger

logger = get_logger(__name__)


def main():
    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument(
        "-nn",
        "--nn_archive_path",
        nargs="+",
        type=str,
        default="",
        help="Path(s) to the NNArchive.",
    )
    arg_parser.add_argument(
        "-m",
        "--model",
        type=str,
        nargs="+",
        default="",
        help="Model from HubAI.",
    )
    arg_parser.add_argument("-all", action="store_true", help="Run all tests")
    arg_parser.add_argument(
        "-p", "--parser", default="", help="Name of the specific parser to test."
    )
    arg_parser.add_argument(
        "--platform",
        type=str,
        default="",
        help="RVC platform to run the tests on. Default is both.",
    )
    arg_parser.add_argument(
        "-v",
        "--depthai-nodes-version",
        default="main",
        help="Branch or release of depthai-nodes. It is used when selecting the models to be tested. Default is main - means main branch and selects all available models. If release tag is provided it will look only for specific set of models.",
    )

    args = arg_parser.parse_args()
    nn_archive_path = args.nn_archive_path  # it is a list of paths
    model = args.model
    run_all = args.all
    parser = args.parser
    rvc_platform = "both" if args.platform == "" else args.platform

    supported_models = json.load(open("supported_models.json"))
    if args.depthai_nodes_version not in supported_models:
        run_all = True
    else:
        run_all = False
        model = supported_models[args.depthai_nodes_version]
        logger.info(
            f"\nRunning on release version {args.depthai_nodes_version} with preselected {len(model)} models."
        )

    if model:
        run_all = False

    logger.info(f"\nRun all tests: {run_all}")
    logger.info(f"RVC2 IP: {os.getenv('RVC2_IP', '')}")
    logger.info(f"RVC4 IP: {os.getenv('RVC4_IP', '')}")
    logger.info(
        f"RVC platform: {'RVC2 & RVC4' if rvc_platform == '' else rvc_platform}"
    )

    if run_all and (nn_archive_path or model):
        raise ValueError("You can't pass both -all and -nn_archive_path or -model")

    if run_all:
        model = get_model_slugs_from_zoo()

    if parser:
        model = find_slugs_from_zoo(parser)
        if len(model) == 0:
            raise ValueError(f"No models found for parser {parser}")
        else:
            logger.info(f"Found models for parser {parser}: {model}")

    if not nn_archive_path and not model:
        raise ValueError("You have to pass either path to NNArchive or model")

    model = [f"{m}" for m in model]

    command = [
        "test_e2e.py",
        f"--nn_archive_path={nn_archive_path}",
        f"--platform={rvc_platform}",
        "-v",
        "--tb=short",
        "-r a",
        "--log-cli-level=DEBUG",
        "--color=yes",
        "-s",
    ]

    if model:
        command = [
            "test_e2e.py",
            f"--model={model}",
            f"--platform={rvc_platform}",
            "-v",
            "--tb=short",
            "-r a",
            "--log-cli-level=DEBUG",
            "--color=yes",
            "-s",
        ]

    exit_code = pytest.main(command)
    sys.exit(exit_code)


if __name__ == "__main__":
    main()



from typing import List

import depthai_nodes.message.creators as creators

from .constants import CLASSIFICATION


def create_classifications(
    classes: List[str] = CLASSIFICATION["classes"],
    scores: List[float] = CLASSIFICATION["scores"],
):
    return creators.create_classification_message(classes=classes, scores=scores)


def create_classifications_sequence(
    classes: List[str] = CLASSIFICATION["classes"],
    scores: List[float] = [
        CLASSIFICATION["scores"],
    ]
    * CLASSIFICATION["sequence_num"],
):
    return creators.create_classification_sequence_message(
        classes=classes, scores=scores
    )



import copy

import cv2
import depthai as dai
import numpy as np
from numpy.typing import NDArray


class SegmentationMask(dai.Buffer):
    """SegmentationMask class for a single- or multi-object segmentation mask.
    Unassigned pixels are represented with "-1" and foreground classes with non-negative
    integers.

    Attributes
    ----------
    mask: NDArray[np.int16]
        Segmentation mask.
    transformation : dai.ImgTransformation
        Image transformation object.
    """

    def __init__(self):
        """Initializes the SegmentationMask object."""
        super().__init__()
        self._mask: NDArray[np.int16] = np.empty(0, dtype=np.int16)
        self._transformation: dai.ImgTransformation = None

    def copy(self):
        """Creates a new instance of the SegmentationMask class and copies the
        attributes.

        @return: A new instance of the SegmentationMask class.
        @rtype: SegmentationMask
        """
        new_obj = SegmentationMask()
        new_obj.mask = copy.deepcopy(self._mask)
        new_obj.setSequenceNum(self.getSequenceNum())
        new_obj.setTimestamp(self.getTimestamp())
        new_obj.setTimestampDevice(self.getTimestampDevice())
        new_obj.setTransformation(self.transformation)
        return new_obj

    @property
    def mask(self) -> NDArray[np.int16]:
        """Returns the segmentation mask.

        @return: Segmentation mask.
        @rtype: NDArray[np.int16]
        """
        return self._mask

    @mask.setter
    def mask(self, value: NDArray[np.int16]):
        """Sets the segmentation mask.

        @param value: Segmentation mask.
        @type value: NDArray[np.int16])
        @raise TypeError: If value is not a numpy array.
        @raise ValueError: If value is not a 2D numpy array.
        @raise ValueError: If each element is not of type int16.
        @raise ValueError: If any element is smaller than -1.
        """
        if not isinstance(value, np.ndarray):
            raise TypeError("Mask must be a numpy array.")
        if not (value.size == 0 or value.ndim == 2):
            raise ValueError("Mask must be 2D or empty.")
        if value.dtype != np.int16:
            raise ValueError("Mask must be an array of int16.")
        if np.any((value < -1)):
            raise ValueError("Mask must be an array of integers larger or equal to -1.")
        self._mask = value

    @property
    def transformation(self) -> dai.ImgTransformation:
        """Returns the Image Transformation object.

        @return: The Image Transformation object.
        @rtype: dai.ImgTransformation
        """
        return self._transformation

    @transformation.setter
    def transformation(self, value: dai.ImgTransformation):
        """Sets the Image Transformation object.

        @param value: The Image Transformation object.
        @type value: dai.ImgTransformation
        @raise TypeError: If value is not a dai.ImgTransformation object.
        """

        if value is not None:
            if not isinstance(value, dai.ImgTransformation):
                raise TypeError(
                    f"Transformation must be a dai.ImgTransformation object, instead got {type(value)}."
                )
        self._transformation = value

    def setTransformation(self, transformation: dai.ImgTransformation):
        """Sets the Image Transformation object.

        @param transformation: The Image Transformation object.
        @type transformation: dai.ImgTransformation
        @raise TypeError: If value is not a dai.ImgTransformation object.
        """
        self.transformation = transformation

    def getVisualizationMessage(self) -> dai.ImgFrame:
        """Returns the default visualization message for segmentation masks."""
        img_frame = dai.ImgFrame()
        mask = self._mask.copy()

        unique_values = np.unique(mask[mask >= 0])
        scaled_mask = np.zeros_like(mask, dtype=np.uint8)

        if unique_values.size == 0:
            return img_frame.setCvFrame(scaled_mask, dai.ImgFrame.Type.BGR888i)

        min_val, max_val = unique_values.min(), unique_values.max()

        if min_val == max_val:
            scaled_mask = np.ones_like(mask, dtype=np.uint8) * 255
        else:
            scaled_mask = ((mask - min_val) / (max_val - min_val) * 255).astype(
                np.uint8
            )
        scaled_mask[mask == -1] = 0
        colored_mask = cv2.applyColorMap(scaled_mask, cv2.COLORMAP_RAINBOW)
        colored_mask[mask == -1] = [0, 0, 0]

        return img_frame.setCvFrame(colored_mask, dai.ImgFrame.Type.BGR888i)



import argparse

import depthai as dai
from utils import get_input_shape, get_num_inputs

from depthai_nodes.logging import get_logger
from depthai_nodes.node import ParsingNeuralNetwork

parser = argparse.ArgumentParser()
parser.add_argument(
    "-nn", "--nn_archive", type=str, default=None, help="Path to the NNArchive."
)
parser.add_argument("-m", "--model", type=str, default=None, help="Model from HubAI.")
parser.add_argument("-ip", type=str, default="", help="IP of the device")
args = parser.parse_args()

logger = get_logger(__name__)

if args.model:
    if "xfeat" in args.model:
        logger.warning("XFeat model is not supported in this test.")
        exit(8)

if not (args.nn_archive or args.model):
    raise ValueError("You have to pass either path to NNArchive or model.")

try:
    device = dai.Device(dai.DeviceInfo(args.ip))
    logger.debug(f"(1) Connected to device with IP/mxid: {args.ip}")
except Exception as e:
    logger.warning(e)
    logger.warning("Can't connect to the device with IP/mxid: %s", args.ip)
    exit(6)

with dai.Pipeline(device) as pipeline:
    camera_node = pipeline.create(dai.node.Camera).build()
    logger.debug("(2) Camera node created.")
    if args.model:
        model_desc = dai.NNModelDescription(
            model=args.model,
            platform=device.getPlatform().name,
        )
        try:
            nn_archive_path = dai.getModelFromZoo(model_desc, useCached=False)
            logger.debug("(3) NN archive downloaded.")
        except Exception as e:
            logger.warning(e)
            logger.warning(
                f"Couldn't find model {args.model} for {device.getPlatform().name} in the ZOO"
            )
            device.close()
            exit(7)
        try:
            nn_archive = dai.NNArchive(nn_archive_path)
            logger.debug("(4) NN archive loaded.")
        except Exception as e:
            logger.debug(e)
            logger.warning(f"Couldn't load the model {args.model} from NN archive.")
            device.close()
            exit(9)

    else:
        nn_archive = dai.NNArchive(args.nn_archive)

    model_platforms = [platform.name for platform in nn_archive.getSupportedPlatforms()]

    if device.getPlatform().name not in model_platforms:
        logger.warning(f"Model not supported on {device.getPlatform().name}.")
        device.close()
        exit(5)

    if get_num_inputs(nn_archive) > 1:
        logger.warning(
            "This model has more than one input. Currently, only models with one input are supported."
        )
        device.close()
        exit(8)

    try:
        input_size = get_input_shape(nn_archive)
    except Exception as e:
        logger.warning(e)
        device.close()
        exit(8)

    image_type = (
        dai.ImgFrame.Type.BGR888i
        if device.getPlatform().name == "RVC4"
        else dai.ImgFrame.Type.BGR888p
    )
    manip = None

    if input_size[0] % 2 != 0 or input_size[1] % 2 != 0:
        manip = pipeline.create(dai.node.ImageManip)
        manip.initialConfig.setOutputSize(input_size[0], input_size[1])
        large_input_shape = (input_size[0] * 2, input_size[1] * 2)

    if input_size[0] < 128 and input_size[1] < 128:
        manip = pipeline.create(dai.node.ImageManip)
        manip.initialConfig.setOutputSize(input_size[0], input_size[1])
        large_input_shape = (input_size[0] * 4, input_size[1] * 4)

    if manip:
        camera_node.requestOutput(large_input_shape, type=image_type, fps=20.0).link(
            manip.inputImage
        )
        nn_w_parser = pipeline.create(ParsingNeuralNetwork).build(manip.out, nn_archive)
        logger.debug("(5) ImageManip node and ParsingNeuralNetwork node created.")
    else:
        nn_w_parser = pipeline.create(ParsingNeuralNetwork).build(
            camera_node, nn_archive, fps=20.0
        )
        logger.debug("(5) ParsingNeuralNetwork node created.")

    head_indices = nn_w_parser._parsers.keys()

    parser_output_queues = {
        i: nn_w_parser.getOutput(i).createOutputQueue() for i in head_indices
    }

    logger.debug("(6) Parser output queues created.")

    pipeline.start()
    logger.debug("(7) Pipeline started.")

    while pipeline.isRunning():
        logger.debug("(8) Pipeline is running.")
        for head_id in parser_output_queues:
            parser_output = parser_output_queues[head_id].get()
            logger.debug(f"Parser output {head_id} - {type(parser_output)}")
        pipeline.stop()

device.close()



import depthai as dai
import numpy as np
import pytest

from depthai_nodes import Classifications


@pytest.fixture
def classifications():
    return Classifications()


def test_initialization(classifications: Classifications):
    assert classifications.classes == []
    assert np.array_equal(classifications.scores, np.array([]))
    assert classifications.transformation is None


def test_set_classes(classifications: Classifications):
    classes = ["cat", "dog", "bird"]
    classifications.classes = classes
    assert classifications.classes == classes

    with pytest.raises(TypeError):
        classifications.classes = "not a list"

    with pytest.raises(ValueError):
        classifications.classes = ["cat", 123, "bird"]


def test_set_scores(classifications: Classifications):
    scores = np.array([0.9, 0.05, 0.05], dtype=np.float32)
    classifications.scores = scores
    assert np.array_equal(classifications.scores, scores)

    with pytest.raises(TypeError):
        classifications.scores = "not an array"

    with pytest.raises(ValueError):
        classifications.scores = np.array([[0.9, 0.05, 0.05]], dtype=np.float32)

    with pytest.raises(ValueError):
        classifications.scores = np.array([0.9, 0.05, "not a float"], dtype=object)


def test_top_class(classifications: Classifications):
    classes = ["cat", "dog", "bird"]
    classifications.classes = classes
    assert classifications.top_class == "cat"


def test_top_score(classifications: Classifications):
    scores = np.array([0.9, 0.05, 0.05], dtype=np.float32)
    classifications.scores = scores
    assert np.allclose(classifications.top_score, 0.9, atol=1e-3)


def test_set_transformation(classifications: Classifications):
    transformation = dai.ImgTransformation()
    classifications.transformation = transformation
    assert classifications.transformation == transformation

    with pytest.raises(TypeError):
        classifications.transformation = "not a dai.ImgTransformation"


def test_set_transformation_none(classifications: Classifications):
    classifications.transformation = None
    assert classifications.transformation is None



import numpy as np


def get_top_values(heatmap):
    """Get the top values from the heatmap tensor.

    @param heatmap: Heatmap tensor.
    @type heatmap: np.ndarray
    @return: Y and X coordinates of the top values.
    @rtype: Tuple[np.ndarray, np.ndarray]
    """
    batchsize, ny, nx, num_joints = heatmap.shape
    heatmap_flat = heatmap.reshape(batchsize, nx * ny, num_joints)

    heatmap_top = np.argmax(heatmap_flat, axis=1)

    Y, X = (heatmap_top // nx), (heatmap_top % nx)
    return Y, X


def get_pose_prediction(heatmap, locref, scale_factors):
    """Get the pose prediction from the heatmap and locref tensors. Used for SuperAnimal
    model.

    @param heatmap: Heatmap tensor.
    @type heatmap: np.ndarray
    @param locref: Locref tensor.
    @type locref: np.ndarray
    @param scale_factors: Scale factors for the x and y axes.
    @type scale_factors: Tuple[float, float]
    @return: Pose prediction.
    @rtype: np.ndarray
    """
    Y, X = get_top_values(heatmap)
    batch_size, num_joints = X.shape

    heatmap_values = heatmap[0, Y.ravel(), X.ravel(), np.arange(num_joints)].reshape(
        1, 1, num_joints, 1
    )  # Shape: [num_joints, 1] for DZ's 3rd dimension

    if locref is not None:
        locref_updates = locref[
            0, Y.ravel(), X.ravel(), np.arange(num_joints), :
        ].reshape(
            1, 1, num_joints, 2
        )  # Shape: [1, 1, num_joints, 2] for DZ's 1st and 2nd dimensions
    else:
        locref_updates = np.zeros((1, 1, num_joints, 2))

    DZ = np.concatenate(
        (locref_updates, heatmap_values), axis=3
    )  # Concatenate along the last dimension to form [1, 1, num_joints, 3]

    X = X * scale_factors[1] + 0.5 * scale_factors[1] + DZ[:, :, :, 0]
    Y = Y * scale_factors[0] + 0.5 * scale_factors[0] + DZ[:, :, :, 1]

    pose = np.concatenate(
        (np.expand_dims(X, axis=-1), np.expand_dims(Y, axis=-1), DZ[:, :, :, 2:3]),
        axis=3,
    )

    return pose



import copy
from typing import List, Tuple

import depthai as dai

from depthai_nodes import PRIMARY_COLOR, SECONDARY_COLOR
from depthai_nodes.logging import get_logger


class Keypoint(dai.Buffer):
    """Keypoint class for storing a keypoint.

    Attributes
    ----------
    x: float
        X coordinate of the keypoint, relative to the input height.
    y: float
        Y coordinate of the keypoint, relative to the input width.
    z: Optional[float]
        Z coordinate of the keypoint.
    confidence: Optional[float]
        Confidence of the keypoint.
    label_name: Optional[str]
        Label name of the keypoint.
    """

    def __init__(self):
        """Initializes the Keypoint object."""
        super().__init__()
        self._x: float = None
        self._y: float = None
        self._z: float = 0.0
        self._confidence: float = -1.0
        self._label_name: str = None
        self._logger = get_logger(__name__)

    def copy(self):
        """Creates a new instance of the Keypoint class and copies the attributes.

        @return: A new instance of the Keypoint class.
        @rtype: Keypoint
        """
        new_obj = Keypoint()
        new_obj.x = copy.deepcopy(self.x)
        new_obj.y = copy.deepcopy(self.y)
        new_obj.z = copy.deepcopy(self.z)
        new_obj.confidence = copy.deepcopy(self.confidence)
        new_obj.label_name = copy.deepcopy(self.label_name)
        return new_obj

    @property
    def x(self) -> float:
        """Returns the X coordinate of the keypoint.

        @return: X coordinate of the keypoint.
        @rtype: float
        """
        return self._x

    @x.setter
    def x(self, value: float):
        """Sets the X coordinate of the keypoint.

        @param value: X coordinate of the keypoint.
        @type value: float
        @raise TypeError: If value is not a float.
        @raise ValueError: If value is not between 0 and 1.
        """
        if not isinstance(value, float):
            raise TypeError("x must be a float.")
        if value < -0.1 or value > 1.1:
            raise ValueError("x must be between 0 and 1.")
        if not (0 <= value <= 1):
            value = max(0, min(1, value))
            self._logger.info("x value was clipped to [0, 1].")
        self._x = value

    @property
    def y(self) -> float:
        """Returns the Y coordinate of the keypoint.

        @return: Y coordinate of the keypoint.
        @rtype: float
        """
        return self._y

    @y.setter
    def y(self, value: float):
        """Sets the Y coordinate of the keypoint.

        @param value: Y coordinate of the keypoint.
        @type value: float
        @raise TypeError: If value is not a float.
        @raise ValueError: If value is not between 0 and 1.
        """
        if not isinstance(value, float):
            raise TypeError("y must be a float.")
        if value < -0.1 or value > 1.1:
            raise ValueError("y must be between 0 and 1.")
        if not (0 <= value <= 1):
            value = max(0, min(1, value))
            self._logger.info("y value was clipped to [0, 1].")
        self._y = value

    @property
    def z(self) -> float:
        """Returns the Z coordinate of the keypoint.

        @return: Z coordinate of the keypoint.
        @rtype: float
        """
        return self._z

    @z.setter
    def z(self, value: float):
        """Sets the Z coordinate of the keypoint.

        @param value: Z coordinate of the keypoint.
        @type value: float
        @raise TypeError: If value is not a float.
        """
        if not isinstance(value, float):
            raise TypeError("z must be a float.")
        self._z = value

    @property
    def confidence(self) -> float:
        """Returns the confidence of the keypoint.

        @return: Confidence of the keypoint.
        @rtype: float
        """
        return self._confidence

    @confidence.setter
    def confidence(self, value: float):
        """Sets the confidence of the keypoint.

        @param value: Confidence of the keypoint.
        @type value: float
        @raise TypeError: If value is not a float.
        @raise ValueError: If value is not between 0 and 1.
        """
        if not isinstance(value, float):
            raise TypeError("confidence must be a float.")
        if (value < -0.1 or value > 1.1) and value != -1.0:
            raise ValueError("Confidence must be between 0 and 1.")
        if not (0 <= value <= 1):
            value = max(0, min(1, value))
            self._logger.info("Confidence value was clipped to [0, 1].")
        self._confidence = value

    @property
    def label_name(self) -> str:
        """Returns the label name of the keypoint.

        @return: Label name of the keypoint.
        @rtype: str
        """
        return self._label_name

    @label_name.setter
    def label_name(self, value: str):
        """Sets the label name of the keypoint.

        @param value: Label name of the keypoint.
        @type value: str
        """
        if value is not None:
            if not isinstance(value, str):
                raise TypeError("label_name must be a string.")
        self._label_name = value


class Keypoints(dai.Buffer):
    """Keypoints class for storing keypoints.

    Attributes
    ----------
    keypoints: List[Keypoint]
        List of Keypoint objects, each representing a keypoint.
    edges: List[Tuple[int, int]]
        List of edges, each representing a connection between two keypoints. NOTE: If you create a Keypoints message with a `create_keypoints_message` function, the edges will be filtered to only include the edges between the keypoints that are present in the filtered keypoints. This is done to ensure that the edges are only drawn between the keypoints that are present in the filtered keypoints. You can always access the full set of edges in the model's NN archive.
    transformation : dai.ImgTransformation
        Image transformation object.
    """

    def __init__(self):
        """Initializes the Keypoints object."""
        super().__init__()
        self._keypoints: List[Keypoint] = []
        self._edges: List[Tuple[int, int]] = []
        self._transformation: dai.ImgTransformation = None

    def copy(self):
        """Creates a new instance of the Keypoints class and copies the attributes.

        @return: A new instance of the Keypoints class.
        @rtype: Keypoints
        """
        new_obj = Keypoints()
        new_obj.keypoints = [keypoint.copy() for keypoint in self.keypoints]
        new_obj.edges = copy.deepcopy(self.edges)
        new_obj.setSequenceNum(self.getSequenceNum())
        new_obj.setTimestamp(self.getTimestamp())
        new_obj.setTimestampDevice(self.getTimestampDevice())
        new_obj.setTransformation(self.transformation)
        return new_obj

    @property
    def keypoints(self) -> List[Keypoint]:
        """Returns the keypoints.

        @return: List of keypoints.
        @rtype: List[Keypoint]
        """
        return self._keypoints

    @keypoints.setter
    def keypoints(self, value: List[Keypoint]):
        """Sets the keypoints.

        @param value: List of keypoints.
        @type value: List[Keypoint]
        @raise TypeError: If value is not a list.
        @raise TypeError: If each each element is not of type Keypoint.
        """
        if not isinstance(value, list):
            raise TypeError("keypoints must be a list.")
        if not all(isinstance(item, Keypoint) for item in value):
            raise ValueError("keypoints must be a list of Keypoint objects.")
        self._keypoints = value

    @property
    def edges(self) -> List[Tuple[int, int]]:
        """Returns the edges.

        @return: List of edges.
        @rtype: List[Tuple[int, int]]
        """
        return self._edges

    @edges.setter
    def edges(self, value: List[Tuple[int, int]]):
        """Sets the edges.

        @param value: List of edges.
        @type value: List[Tuple[int, int]]
        @raise TypeError: If value is not a list.
        @raise TypeError: If each each element is not of type Tuple[int, int].
        """
        if not isinstance(value, list):
            raise TypeError("edges must be a list.")
        if not all(
            (isinstance(item, tuple) or isinstance(item, list))
            and len(item) == 2
            and all(isinstance(i, int) for i in item)
            for item in value
        ):
            raise TypeError("edges must be a list of tuples or lists of integers.")
        self._edges = value

    @property
    def transformation(self) -> dai.ImgTransformation:
        """Returns the Image Transformation object.

        @return: The Image Transformation object.
        @rtype: dai.ImgTransformation
        """
        return self._transformation

    @transformation.setter
    def transformation(self, value: dai.ImgTransformation):
        """Sets the Image Transformation object.

        @param value: The Image Transformation object.
        @type value: dai.ImgTransformation
        @raise TypeError: If value is not a dai.ImgTransformation object.
        """

        if value is not None:
            if not isinstance(value, dai.ImgTransformation):
                raise TypeError(
                    f"Transformation must be a dai.ImgTransformation object, instead got {type(value)}."
                )
        self._transformation = value

    def setTransformation(self, transformation: dai.ImgTransformation):
        """Sets the Image Transformation object.

        @param transformation: The Image Transformation object.
        @type transformation: dai.ImgTransformation
        @raise TypeError: If value is not a dai.ImgTransformation object.
        """
        self.transformation = transformation

    def getPoints2f(self) -> dai.VectorPoint2f:
        """Returns the keypoints in the form of a dai.VectorPoint2f object."""

        return dai.VectorPoint2f(
            [dai.Point2f(keypoint.x, keypoint.y) for keypoint in self.keypoints]
        )

    def getPoints3f(self) -> List[dai.Point3f]:
        """Returns the keypoints in the form of a list of dai.Point3f objects."""

        return [
            dai.Point3f(keypoint.x, keypoint.y, keypoint.z)
            for keypoint in self.keypoints
        ]

    def getVisualizationMessage(self) -> dai.ImgAnnotations:
        """Creates a default visualization message for the keypoints."""
        img_annotations = dai.ImgAnnotations()
        annotation = dai.ImgAnnotation()

        pointsAnnotation = dai.PointsAnnotation()
        pointsAnnotation.type = dai.PointsAnnotationType.POINTS
        pointsAnnotation.points = self.getPoints2f()
        pointsAnnotation.outlineColor = PRIMARY_COLOR
        pointsAnnotation.fillColor = PRIMARY_COLOR
        pointsAnnotation.thickness = 1
        annotation.points.append(pointsAnnotation)

        for edge in self.edges:
            pt1_ix, pt2_ix = edge
            pt1 = self.keypoints[pt1_ix]
            pt2 = self.keypoints[pt2_ix]
            pointsAnnotation = dai.PointsAnnotation()
            pointsAnnotation.type = dai.PointsAnnotationType.LINE_STRIP
            pointsAnnotation.points = dai.VectorPoint2f(
                [dai.Point2f(pt1.x, pt1.y), dai.Point2f(pt2.x, pt2.y)]
            )
            pointsAnnotation.outlineColor = SECONDARY_COLOR
            pointsAnnotation.fillColor = SECONDARY_COLOR
            pointsAnnotation.thickness = 1
            annotation.points.append(pointsAnnotation)

        img_annotations.annotations.append(annotation)
        img_annotations.setTimestamp(self.getTimestamp())
        return img_annotations



from abc import ABCMeta, abstractmethod
from typing import Any, Dict

import depthai as dai

from depthai_nodes.logging import get_logger


class BaseMeta(ABCMeta, type(dai.node.ThreadedHostNode)):
    pass


class BaseParser(dai.node.ThreadedHostNode, metaclass=BaseMeta):
    """Base class for neural network output parsers. This class serves as a foundation
    for specific parser implementations used to postprocess the outputs of neural
    network models. Each parser is attached to a model "head" that governs the parsing
    process as it contains all the necessary information for the parser to function
    correctly. Subclasses should implement `build` method to correctly set all
    parameters of the parser and the `run` method to define the parsing logic.

    Attributes
    ----------
    input : Node.Input
        Node's input. It is a linking point to which the Neural Network's output is linked. It accepts the output of the Neural Network node.
    out : Node.Output
        Parser sends the processed network results to this output in a form of DepthAI message. It is a linking point from which the processed network results are retrieved.
    """

    def __init__(self) -> None:
        super().__init__()
        self._input = self.createInput()
        self._out = self.createOutput()
        self._logger = get_logger(self.__class__.__name__)

    @property
    @abstractmethod
    def input(self) -> dai.Node.Input:
        return self._input

    @property
    @abstractmethod
    def out(self) -> dai.Node.Output:
        return self._out

    @input.setter
    def input(self, node: dai.Node.Input) -> None:
        """Linking point to which the Neural Network's output is linked."""
        self._input = node

    @out.setter
    def out(self, node: dai.Node.Output) -> None:
        """Output node to which the processed network results are sent in the form of a
        DepthAI message."""
        self._out = node

    @abstractmethod
    def build(self, head_config: Dict[str, Any]) -> "BaseParser":
        """Configures the parser based on the specified head configuration.

        @param head_config: A dictionary containing configuration details relevant to
            the parser, including parameters and settings required for output parsing.
        @type head_config: Dict[str, Any]
        @return: The parser object with the head configuration set.
        @rtype: BaseParser
        """
        pass

    @abstractmethod
    def run(self):
        """Parses the output from the neural network head.

        This method should be overridden by subclasses to implement the specific parsing
        logic. It accepts arbitrary keyword arguments for flexibility.

        @param kwargs: Arbitrary keyword arguments for the parsing process.
        @type kwargs: Any
        @return message: The parsed output message, as defined by the logic in the
            subclass.
        @rtype message: Any
        """
        pass



from depthai_nodes.node.parsing_neural_network import ParsingNeuralNetwork


class HostParsingNeuralNetwork(ParsingNeuralNetwork):
    def _generateParsers(self, parserGenerator, nnArchive):
        return parserGenerator.build(nnArchive, host_only=True)
